import logging
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import List, Optional

from spacepackets.cfdp import ConditionCode, FileStoreResponseTlv, MessageToUserTlv
from spacepackets.cfdp.pdu.file_data import RecordContinuationState
from spacepackets.cfdp.pdu.finished import DeliveryCode, FileDeliveryStatus
from spacepackets.util import UnsignedByteField
from tmtccmd.cfdp.defs import TransactionId
from tmtccmd.cfdp.filestore import VirtualFilestore, HostFilestore

_LOGGER = logging.getLogger(__name__)


@dataclass
class MetadataRecvParams:
    transaction_id: TransactionId
    source_id: UnsignedByteField
    file_size: Optional[int]
    source_file_name: Optional[str]
    dest_file_name: Optional[str]
    msgs_to_user: Optional[List[MessageToUserTlv]] = None


@dataclass
class TransactionFinishedParams:
    transaction_id: TransactionId
    condition_code: ConditionCode
    file_status: FileDeliveryStatus
    delivery_code: DeliveryCode
    fs_responses: Optional[List[FileStoreResponseTlv]] = None
    status_report: Optional[any] = None


@dataclass
class FileSegmentRecvdParams:
    """The length of the segment metadata is not supplied as an extra parameter as it can be
    simply queried with len(segment_metadata)
    """

    transaction_id: TransactionId
    offset: int
    length: int
    record_cont_state: Optional[RecordContinuationState]
    segment_metadata: Optional[bytes]


class CfdpUserBase(ABC):
    """This user base class provides the primary user interface to interact with CFDP handlers.
    It is also used to pass the Virtual Filestore (VFS) implementation to the CFDP handlers
    so the filestore operations can be mapped to the underlying filestore.

    This class is used by implementing it in a child class and then passing it to the CFDP
    handler objects. The base class provides default implementation for the user indication
    primitives specified in the CFDP standard. The user can override these implementations
    to provide custom indication handlers.
    """

    def __init__(self, vfs: Optional[VirtualFilestore] = None):
        if vfs is None:
            vfs = HostFilestore()
        self.vfs = vfs

    @abstractmethod
    def transaction_indication(self, transaction_id: TransactionId):
        """This indication is used to report the transaction ID to the CFDP user"""
        _LOGGER.info(f"Transaction.indication for {transaction_id}")

    @abstractmethod
    def eof_sent_indication(self, transaction_id: TransactionId):
        _LOGGER.info(f"EOF-Sent.indication for {transaction_id}")

    @abstractmethod
    def transaction_finished_indication(self, params: TransactionFinishedParams):
        _LOGGER.info(
            f"Transaction-Finished.indication for {params.transaction_id}. Parameters:"
        )
        print(params)

    @abstractmethod
    def metadata_recv_indication(self, params: MetadataRecvParams):
        _LOGGER.info(
            f"Metadata-Recv.indication for {params.transaction_id}. Parameters:"
        )
        print(params)

    @abstractmethod
    def file_segment_recv_indication(self, params: FileSegmentRecvdParams):
        _LOGGER.info(
            f"File-Segment-Recv.indication for {params.transaction_id}. Parameters:"
        )
        print(params)

    @abstractmethod
    def report_indication(self, transaction_id: TransactionId, status_report: any):
        # TODO: p.28 of the CFDP standard specifies what information the status report parameter
        #       could contain. I think it would be better to not hardcode the type of the status
        #       report here, but something like Union[any, CfdpStatusReport] with CfdpStatusReport
        #       being an implementation which supports all three information suggestions would be
        #       nice
        pass

    @abstractmethod
    def suspended_indication(
        self, transaction_id: TransactionId, cond_code: ConditionCode
    ):
        _LOGGER.info(
            f"Suspended.indication for {transaction_id} | Condition Code: {cond_code}"
        )

    @abstractmethod
    def resumed_indication(self, transaction_id: TransactionId, progress: int):
        _LOGGER.info(
            f"Resumed.indication for {transaction_id} | Progress: {progress} bytes"
        )

    @abstractmethod
    def fault_indication(
        self, transaction_id: TransactionId, cond_code: ConditionCode, progress: int
    ):
        _LOGGER.warning(
            f"Fault.indication for {transaction_id} | Condition Code: {cond_code} | "
            f"Progress: {progress} bytes"
        )

    @abstractmethod
    def abandoned_indication(
        self, transaction_id: TransactionId, cond_code: ConditionCode, progress: int
    ):
        _LOGGER.warning(
            f"Abandoned.indication for {transaction_id} | Condition Code: {cond_code} | "
            f"Progress: {progress} bytes"
        )

    @abstractmethod
    def eof_recv_indication(self, transaction_id: TransactionId):
        _LOGGER.info(f"EOF-Recv.indication for {transaction_id}")



from collections import deque
from datetime import timedelta
from typing import cast
from unittest import TestCase
from unittest.mock import MagicMock

from spacepackets.ecss import PusTelecommand
from tmtccmd import DefaultProcedureInfo
from tmtccmd.tc import WaitEntry, QueueEntryHelper

# Required for eval calls
# noinspection PyUnresolvedReferences
from tmtccmd.tc import LogQueueEntry, RawTcEntry  # noqa: F401
from tmtccmd.tc.queue import QueueWrapper, DefaultPusQueueHelper
from tmtccmd.util import ProvidesSeqCount


class TestTcQueue(TestCase):
    def setUp(self) -> None:
        self.queue_wrapper = QueueWrapper(
            info=DefaultProcedureInfo.empty(), queue=deque()
        )
        self.assertEqual(self.queue_wrapper.queue, deque())
        self.queue_helper = DefaultPusQueueHelper(
            self.queue_wrapper,
            tc_sched_timestamp_len=4,
            seq_cnt_provider=None,
            default_pus_apid=None,
            pus_verificator=None,
        )
        self.pus_cmd = PusTelecommand(service=17, subservice=1)

    def test_wait_entry(self):
        self.queue_helper.add_wait(timedelta(seconds=2))
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        wait_entry = cast(WaitEntry, self.queue_wrapper.queue.pop())
        self.assertTrue(wait_entry)
        self.assertFalse(wait_entry.is_tc())
        self.assertEqual(wait_entry.wait_time.total_seconds(), 2.0)
        self.assertEqual(len(self.queue_wrapper.queue), 0)

    def test_apid_stamper(self):
        self.queue_helper.pus_apid = 0x22
        self.queue_helper.add_pus_tc(self.pus_cmd)
        cast_wrapper = QueueEntryHelper(self.queue_wrapper.queue.popleft())
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, self.pus_cmd)
        self.assertTrue(pus_entry)
        self.assertEqual(pus_entry.pus_tc.apid, 0x22)

    def test_seq_count_stamper(self):
        self.queue_helper.seq_cnt_provider = MagicMock(spec=ProvidesSeqCount)
        self.queue_helper.seq_cnt_provider.get_and_increment.return_value = 5
        self.queue_helper.add_pus_tc(self.pus_cmd)
        cast_wrapper = QueueEntryHelper(self.queue_wrapper.queue.popleft())
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc.seq_count, 5)

    def test_faulty_cast(self):
        self.queue_helper.add_pus_tc(self.pus_cmd)
        cast_wrapper = QueueEntryHelper(self.queue_wrapper.queue.popleft())
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, self.pus_cmd)
        self.assertTrue(pus_entry)
        with self.assertRaises(TypeError):
            cast_wrapper.to_wait_entry()

    def test_multi_entry(self):
        pus_cmd = PusTelecommand(service=17, subservice=1)
        self.queue_helper.add_pus_tc(pus_cmd)
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        self.queue_helper.add_log_cmd("Test String")
        self.queue_helper.add_raw_tc(bytes([0, 1, 2]))
        self.queue_helper.add_ccsds_tc(pus_cmd.to_space_packet())
        self.queue_helper.add_packet_delay(timedelta(seconds=3.0))
        self.assertEqual(len(self.queue_wrapper.queue), 5)
        pus_entry = self.queue_wrapper.queue.popleft()
        cast_wrapper = QueueEntryHelper(pus_entry)
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, pus_cmd)
        self.assertTrue(pus_entry)
        log_entry = self.queue_wrapper.queue.popleft()
        self.assertFalse(log_entry.is_tc())
        cast_wrapper.base = log_entry
        log_entry = cast_wrapper.to_log_entry()
        self.assertTrue(log_entry)
        with self.assertRaises(TypeError):
            cast_wrapper.to_raw_tc_entry()
        self.assertEqual(log_entry.log_str, "Test String")
        test_obj = eval(f"{log_entry!r}")
        self.assertEqual(test_obj.log_str, log_entry.log_str)

        raw_entry = self.queue_wrapper.queue.popleft()
        self.assertTrue(raw_entry.is_tc())
        cast_wrapper.base = raw_entry
        raw_entry = cast_wrapper.to_raw_tc_entry()
        self.assertTrue(raw_entry)
        self.assertEqual(raw_entry.tc, bytes([0, 1, 2]))
        test_obj = eval(f"{raw_entry!r}")
        self.assertEqual(raw_entry.tc, test_obj.tc)

        space_packet_entry = self.queue_wrapper.queue.popleft()
        self.assertTrue(space_packet_entry.is_tc())
        cast_wrapper.base = space_packet_entry
        space_packet_entry = cast_wrapper.to_space_packet_entry()
        self.assertTrue(space_packet_entry)
        self.assertTrue(space_packet_entry.space_packet, pus_cmd.to_space_packet())

        packet_delay = self.queue_wrapper.queue.pop()
        self.assertFalse(packet_delay.is_tc())
        cast_wrapper.base = packet_delay
        packet_delay = cast_wrapper.to_packet_delay_entry()
        self.assertEqual(packet_delay.delay_time.total_seconds(), 3.0)



from spacepackets.cfdp import ConditionCode
from tmtccmd.cfdp.mib import DefaultFaultHandlerBase


class FaultHandler(DefaultFaultHandlerBase):
    def __init__(self):
        super().__init__()
        self.notice_of_suspension_called = False
        self.notice_of_suspension_call_count = 0

    def notice_of_suspension_cb(self, cond: ConditionCode):
        self.notice_of_suspension_called = True
        self.notice_of_suspension_call_count += 1

    def notice_of_cancellation_cb(self, cond: ConditionCode):
        pass

    def abandoned_cb(self, cond: ConditionCode):
        pass

    def ignore_cb(self, cond: ConditionCode):
        pass



import enum
from typing import Tuple, Dict, Union

# Com Interface Types
ComIfValueT = Tuple[str, any]
ComIfDictT = Dict[str, ComIfValueT]


def default_json_path() -> str:
    return "tmtc_conf.json"


class CoreComInterfaces(str, enum.Enum):
    value: str

    DUMMY = "dummy"
    UDP = "udp"
    TCP = "tcp"
    SERIAL_FIXED_FRAME = "serial_fixed"
    SERIAL_COBS = "serial_cobs"
    SERIAL_DLE = "serial_dle"
    SERIAL_QEMU = "serial_qemu"
    UNSPECIFIED = "unspec"


CORE_COM_IF_DICT = {
    CoreComInterfaces.DUMMY: ("Dummy Interface", None),
    CoreComInterfaces.SERIAL_DLE: ("Serial Interace with DLE encoding", None),
    CoreComInterfaces.UDP: ("TCP/IP with UDP datagrams", None),
    CoreComInterfaces.TCP: ("TCP/IP with TCP", None),
    CoreComInterfaces.SERIAL_FIXED_FRAME: (
        "Serial Interface with fixed size frames",
        None,
    ),
    CoreComInterfaces.SERIAL_QEMU: ("Serial Interface using QEMU", None),
    CoreComInterfaces.UNSPECIFIED: ("Unspecified", None),
}


# Mode options, set by args parser
class CoreModeList(enum.IntEnum):
    """These are the core modes which will be translated to different TC and TM modes
    for the CCSDS backend

    1. ONE_QUEUE_MODE: This mode is optimized to handle one queue. It will configure the backend
       to request program termination upon finishing the queue handling. This is also the
       appropriate solution for single commands where the queue only consists of one telecommand.
    2. LISTENER_MODE: Only listen to TM
    3. MULTI_INTERACTIVE_QUEUE_MODE:
    """

    #
    ONE_QUEUE_MODE = 0
    LISTENER_MODE = 1
    # This mode is optimized for the handling of multiple queues. It will configure the backend
    # to request additional queues or a mode change from the user instead of requesting program
    # termination
    MULTI_INTERACTIVE_QUEUE_MODE = 3
    # The program will not do anything in this mode. This includes polling TM and sending any TCs
    IDLE = 5


class CoreModeConverter:
    @staticmethod
    def get_str(mode: Union[CoreModeList, int]) -> str:
        if mode == CoreModeList.LISTENER_MODE:
            return "listener"
        elif mode == CoreModeList.ONE_QUEUE_MODE:
            return "one-q"
        elif mode == CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE:
            return "multi-q"
        elif mode == CoreModeList.IDLE:
            return "idle"


class CoreServiceList(str, enum.Enum):
    value: str
    SERVICE_2 = "2"
    SERVICE_3 = "3"
    SERVICE_5 = "5"
    SERVICE_8 = "8"
    SERVICE_9 = "9"
    SERVICE_11 = "11"
    SERVICE_17 = "17"
    SERVICE_17_ALT = "test"
    SERVICE_20 = "20"
    SERVICE_23 = "23"
    SERVICE_200 = "200"


DEFAULT_APID = 0xEF
DEBUG_MODE = False



import time
from collections import deque
from typing import cast
from unittest import TestCase
from datetime import timedelta
from unittest.mock import MagicMock, ANY

from spacepackets.ecss import PusTelecommand
from tmtccmd.com import ComInterface
from tmtccmd.tc.ccsds_seq_sender import SequentialCcsdsSender, SenderMode
from tmtccmd.tc.handler import TcHandlerBase, SendCbParams
from tmtccmd.tc.queue import QueueWrapper, DefaultPusQueueHelper


class TestSendReceive(TestCase):
    def setUp(self) -> None:
        self.queue_wrapper = QueueWrapper(info=None, queue=deque())
        self.queue_helper = DefaultPusQueueHelper(
            self.queue_wrapper,
            tc_sched_timestamp_len=4,
            pus_verificator=None,
            default_pus_apid=None,
            seq_cnt_provider=None,
        )
        self.tc_handler_mock = MagicMock(spec=TcHandlerBase)
        self.com_if = MagicMock(spec=ComInterface)
        self.seq_sender = SequentialCcsdsSender(
            self.queue_wrapper, self.tc_handler_mock
        )

    def test_basic(self):
        res = self.seq_sender.operation(self.com_if)
        # Queue is empty initially
        self.assertEqual(res.mode, SenderMode.DONE)
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)
        self.assertTrue(self.seq_sender.no_delay_remaining())
        self.queue_helper.add_raw_tc(bytes([0, 1, 2]))
        # One queue entry which should be handled immediately
        self.seq_sender.queue_wrapper = self.queue_wrapper
        self.assertEqual(self.seq_sender.mode, SenderMode.BUSY)
        # Is busy now, so does not accept new queue unless forced
        with self.assertRaises(ValueError):
            self.seq_sender.queue_wrapper = self.queue_wrapper
        self.seq_sender.operation(self.com_if)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertIsNone(send_cb_params.info.base)
        raw_tc_entry = send_cb_params.entry.to_raw_tc_entry()
        self.assertEqual(raw_tc_entry.tc, bytes([0, 1, 2]))
        # Queue should be empty now
        # Called twice for each operation call
        self.assertEqual(self.tc_handler_mock.queue_finished_cb.call_count, 2)
        self.assertFalse(self.queue_wrapper.queue)
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)
        self.queue_helper.add_raw_tc(bytes([3, 2, 1]))
        self.seq_sender.resume()
        self.assertEqual(self.seq_sender.mode, SenderMode.BUSY)
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([3, 2, 1]))

    def test_with_wait_entry(self):
        wait_delay = 0.01
        self.queue_helper.add_raw_tc(bytes([3, 2, 1]))
        self.queue_helper.add_wait(timedelta(seconds=wait_delay))
        self.queue_helper.add_raw_tc(bytes([1, 2, 3]))
        # Resume call necessary
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)
        self.seq_sender.resume()
        res = self.seq_sender.operation(self.com_if)
        self.assertEqual(res.mode, SenderMode.BUSY)
        self.assertTrue(res.tc_sent)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([3, 2, 1]))
        self.assertTrue(self.seq_sender.no_delay_remaining())
        # 2 queue entries remaining
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 2)
        # Now the wait entry should be handled
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(self.seq_sender.no_delay_remaining())
        self.assertFalse(res.tc_sent)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(
            send_cb_params.entry.to_wait_entry().wait_time,
            timedelta(seconds=wait_delay),
        )
        # Now no TCs should be sent for 10 ms
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 1)
        self.assertEqual(res.mode, SenderMode.BUSY)
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(res.tc_sent)
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 1)
        # After a delay, TC should be sent
        time.sleep(wait_delay)
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 0)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([1, 2, 3]))

    def test_interpacket_delay(self):
        delay_ms = 20
        inter_packet_delay = timedelta(milliseconds=delay_ms)
        ping_cmd = PusTelecommand(service=17, subservice=1)
        self.queue_helper.add_pus_tc(ping_cmd)
        self.queue_helper.add_packet_delay_ms(delay_ms)
        self.queue_helper.add_ccsds_tc(ping_cmd.to_space_packet())
        self.queue_helper.add_raw_tc(bytes([0, 1, 2]))
        # Send first TC, assert delay of 10 ms, then send last packet
        res = self.seq_sender.operation(self.com_if)
        self.assertEqual(res.longest_rem_delay, timedelta())
        self.assertTrue(res.tc_sent)
        self.assertEqual(res.next_entry_is_tc, False)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.com_if, self.com_if)
        self.assertIsNone(send_cb_params.info.base)
        self.assertEqual(send_cb_params.entry.to_pus_tc_entry().pus_tc, ping_cmd)
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(res.tc_sent)
        self.assertEqual(res.next_entry_is_tc, True)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertIsNone(send_cb_params.info.base)
        self.assertEqual(
            send_cb_params.entry.to_packet_delay_entry().delay_time.total_seconds(),
            inter_packet_delay.total_seconds(),
        )
        self.assertTrue(
            inter_packet_delay * 0.8 <= res.longest_rem_delay <= inter_packet_delay
        )
        res = self.seq_sender.operation(self.com_if)
        # No TC sent
        self.assertFalse(res.tc_sent)
        self.assertEqual(res.next_entry_is_tc, True)
        self.assertEqual(len(self.queue_wrapper.queue), 2)
        time.sleep(inter_packet_delay.total_seconds())
        res = self.seq_sender.operation(self.com_if)
        # TC sent
        self.assertTrue(res.tc_sent)
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        res = self.seq_sender.operation(self.com_if)
        # No TC sent, delay after each packet
        self.assertFalse(res.tc_sent)
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        self.assertTrue(
            0.8 * inter_packet_delay < res.longest_rem_delay <= inter_packet_delay
        )
        # Delay 10 ms
        time.sleep(inter_packet_delay.total_seconds())
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        # Queue is empty now, but this should still be set to False
        self.assertEqual(res.next_entry_is_tc, False)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([0, 1, 2]))

    def test_delay_at_end(self):
        delay_at_end = timedelta(milliseconds=20)
        self.queue_helper.add_raw_tc(bytes([3, 2, 1]))
        self.queue_helper.add_wait(delay_at_end)
        self.seq_sender.resume()
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        self.assertEqual(res.longest_rem_delay, timedelta())
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(res.tc_sent)
        self.assertFalse(self.seq_sender.no_delay_remaining())
        self.assertTrue(0.8 * delay_at_end < res.longest_rem_delay <= delay_at_end)
        self.assertEqual(self.seq_sender.mode, SenderMode.BUSY)
        time.sleep(delay_at_end.total_seconds())
        self.assertTrue(self.seq_sender.no_delay_remaining())
        self.seq_sender.operation(self.com_if)
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)



from dataclasses import dataclass
from typing import Sequence, Optional, Tuple

from spacepackets import SpacePacket, SpacePacketHeader, PacketType
from spacepackets.cfdp import GenericPduPacket, PduType, DirectiveType, PduFactory
from spacepackets.cfdp.pdu import PduHolder

from tmtccmd.util import ProvidesSeqCount
from tmtccmd.cfdp import (
    LocalEntityCfg,
    RemoteEntityCfgTable,
    CfdpUserBase,
    RemoteEntityCfg,
)
from tmtccmd.cfdp.request import PutRequest
from tmtccmd.cfdp.defs import CfdpStates
from .defs import NoRemoteEntityCfgFound, BusyError

from .dest import DestStateWrapper
from .dest import DestHandler
from .source import SourceHandler, SourceStateWrapper, FsmResult
from .source import TransactionStep as SourceTransactionStep
from .dest import TransactionStep as DestTransactionStep


@dataclass
class StateWrapper:
    source_handler_state = SourceStateWrapper()
    dest_handler_state = DestStateWrapper()


class CfdpHandler:
    def __init__(
        self,
        cfg: LocalEntityCfg,
        user: CfdpUserBase,
        seq_cnt_provider: ProvidesSeqCount,
        remote_cfgs: Sequence[RemoteEntityCfg],
    ):
        self.remote_cfg_table = RemoteEntityCfgTable()
        self.remote_cfg_table.add_configs(remote_cfgs)
        self.dest_handler = DestHandler(cfg, user, self.remote_cfg_table)
        self.source_handler = SourceHandler(cfg, seq_cnt_provider, user)

    def put_request(self, request: PutRequest):
        if not self.remote_cfg_table.get_cfg(request.cfg.destination_id):
            raise ValueError(
                f"No remote CFDP config found for entity ID {request.cfg.destination_id}"
            )
        self.source_handler.put_request(
            request, self.remote_cfg_table.get_cfg(request.cfg.destination_id)
        )

    def pull_next_source_packet(self) -> Optional[PduHolder]:
        res = self.source_handler.state_machine()
        if res.states.packet_ready:
            return self.source_handler.pdu_holder
        return None

    def pull_next_dest_packet(self) -> Optional[PduHolder]:
        res = self.dest_handler.state_machine()
        if res.states.packet_ready:
            return self.dest_handler.pdu_holder
        return None

    def __iter__(self):
        return self

    def __next__(self) -> (Optional[PduHolder], Optional[PduHolder]):
        """The iterator for this class will returns a tuple of optional PDUs wrapped b a
        :py:class:`PduHolder`.

        :return: Can be a tuple where the first entry can hold a source packet and the second entry
            can be a destination packet. If both packets are None, a StopIteration will be raised.
        """
        next_source_packet = self.pull_next_source_packet()
        next_dest_packet = self.pull_next_dest_packet()
        if not next_dest_packet and not next_source_packet:
            raise StopIteration
        return next_source_packet, next_dest_packet

    def put_request_pending(self) -> bool:
        return self.source_handler.states.state != CfdpStates.IDLE

    def confirm_dest_packet_sent(self):
        self.dest_handler.confirm_packet_sent_advance_fsm()

    def confirm_source_packet_sent(self):
        self.source_handler.confirm_packet_sent_advance_fsm()

    def pass_packet(self, packet: GenericPduPacket):
        """This function routes the packets based on PDU type and directive type if applicable.

        The routing is based on section 4.5 of the CFDP standard whcih specifies the PDU forwarding
        procedure.
        """
        if packet.pdu_type == PduType.FILE_DATA:
            self.dest_handler.pass_packet(packet)
        else:
            if packet.directive_type in [
                DirectiveType.METADATA_PDU,
                DirectiveType.EOF_PDU,
                DirectiveType.PROMPT_PDU,
            ]:
                # Section b) of 4.5.3: These PDUs should always be targeted towards the file
                # receiver a.k.a. the destination handler
                self.dest_handler.pass_packet(packet)
            elif packet.directive_type in [
                DirectiveType.FINISHED_PDU,
                DirectiveType.NAK_PDU,
                DirectiveType.KEEP_ALIVE_PDU,
            ]:
                # Section c) of 4.5.3: These PDUs should always be targeted towards the file sender
                # a.k.a. the source handler
                self.source_handler.pass_packet(packet)
            elif packet.directive_type == DirectiveType.ACK_PDU:
                # Section a): Recipient depends on the type of PDU that is being acknowledged.
                # We can simply extract the PDU type from the raw stream. If it is an EOF PDU,
                # this packet is passed to the source handler. For a finished PDU, it is
                # passed to the destination handler
                pdu_holder = PduHolder(packet)
                ack_pdu = pdu_holder.to_ack_pdu()
                if ack_pdu.directive_code_of_acked_pdu == DirectiveType.EOF_PDU:
                    self.source_handler.pass_packet(packet)
                elif ack_pdu.directive_code_of_acked_pdu == DirectiveType.FINISHED_PDU:
                    self.dest_handler.pass_packet(packet)


class CfdpInCcsdsHandler:
    def __init__(
        self,
        cfg: LocalEntityCfg,
        user: CfdpUserBase,
        remote_cfgs: Sequence[RemoteEntityCfg],
        ccsds_apid: int,
        cfdp_seq_cnt_provider: ProvidesSeqCount,
        ccsds_seq_cnt_provider: ProvidesSeqCount,
    ):
        """Wrapper helper type used to wrap PDU packets into CCSDS packets and to extract PDU
        packets from CCSDS packets.

        :param cfg: Local CFDP entity configuration.
        :param user: User wrapper. This contains the indication callback implementations and the
            virtual filestore implementation.
        :param cfdp_seq_cnt_provider: Every CFDP file transfer has a transaction sequence number.
            This provider is used to retrieve that sequence number.
        :param ccsds_seq_cnt_provider: Each CFDP PDU is wrapped into a CCSDS space packet, and each
            space packet has a dedicated sequence count. This provider is used to retrieve the
            sequence count.
        :param ccsds_apid: APID to use for the CCSDS space packet header wrapped around each PDU.
            This is important so that the OBSW can distinguish between regular PUS packets and
            CFDP packets.
        """
        self.cfdp_handler = CfdpHandler(cfg, user, cfdp_seq_cnt_provider, remote_cfgs)
        self.ccsds_seq_cnt_provider = ccsds_seq_cnt_provider
        self.ccsds_apid = ccsds_apid

    def put_request_pending(self):
        return self.cfdp_handler.put_request_pending()

    def fsm(self):
        self.source_handler.state_machine()
        self.dest_handler.state_machine()

    @property
    def source_handler(self):
        return self.cfdp_handler.source_handler

    @property
    def dest_handler(self):
        return self.cfdp_handler.dest_handler

    def pull_next_source_packet(self) -> Optional[Tuple[PduHolder, SpacePacket]]:
        """Retrieves the next PDU to send and wraps it into a space packet"""
        next_packet = self.cfdp_handler.pull_next_source_packet()
        if next_packet is None:
            return next_packet
        sp_header = SpacePacketHeader(
            packet_type=PacketType.TC,
            apid=self.ccsds_apid,
            seq_count=self.ccsds_seq_cnt_provider.get_and_increment(),
            data_len=next_packet.packet_len - 1,
        )
        return next_packet, SpacePacket(sp_header, None, next_packet.pack())

    def pull_next_dest_packet(self) -> Optional[Tuple[PduHolder, SpacePacket]]:
        """Retrieves the next PDU to send and wraps it into a space packet"""
        next_packet = self.cfdp_handler.pull_next_dest_packet()
        if next_packet is None:
            return next_packet
        sp_header = SpacePacketHeader(
            packet_type=PacketType.TC,
            apid=self.ccsds_apid,
            seq_count=self.ccsds_seq_cnt_provider.get_and_increment(),
            data_len=next_packet.packet_len - 1,
        )
        return next_packet, SpacePacket(sp_header, None, next_packet.pack())

    def confirm_dest_packet_sent(self):
        self.cfdp_handler.confirm_dest_packet_sent()

    def confirm_source_packet_sent(self):
        self.cfdp_handler.confirm_source_packet_sent()

    def pass_space_packet(self, space_packet: SpacePacket):
        # Unwrap the user data and pass it to the handler
        pdu_raw = space_packet.user_data
        pdu_base = PduFactory.from_raw(pdu_raw)
        self.pass_pdu_packet(pdu_base)

    def pass_pdu_packet(self, pdu_base: GenericPduPacket):
        self.cfdp_handler.pass_packet(pdu_base)

    def __iter__(self):
        return self

    def __next__(
        self,
    ) -> (
        Optional[Tuple[PduHolder, SpacePacket]],
        Optional[Tuple[PduHolder, SpacePacket]],
    ):
        """The iterator for this class will returns a tuple of optional PDUs wrapped b a
        :py:class:`PduHolder`.

        :return: Can be a tuple where the first entry can hold a source packet and the second entry
            can be a destination packet. If both packets are None, a StopIteration will be raised.
        """
        next_source_tuple = self.pull_next_source_packet()
        next_dest_tuple = self.pull_next_dest_packet()
        if not next_source_tuple and not next_dest_tuple:
            raise StopIteration
        return next_source_tuple, next_dest_tuple



from .obj_id import ObjectIdU32, ObjectIdU16, ObjectIdU8, ObjectIdBase, ObjectIdDictT
from .retval import RetvalDictT
from .seqcnt import (
    FileSeqCountProvider,
    PusFileSeqCountProvider,
    ProvidesSeqCount,
    SeqCountProvider,
)



from __future__ import annotations

import dataclasses
import enum
import logging
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Deque, cast, Optional

from spacepackets.cfdp import (
    PduType,
    ChecksumType,
    TransmissionMode,
    ConditionCode,
    TlvTypes,
    PduConfig,
    Direction,
    FaultHandlerCode,
)
from spacepackets.cfdp.pdu import (
    DirectiveType,
    AbstractFileDirectiveBase,
    MetadataPdu,
    FileDataPdu,
    EofPdu,
    FinishedPdu,
)
from spacepackets.cfdp.pdu.finished import (
    FinishedParams,
    DeliveryCode,
    FileDeliveryStatus,
)
from spacepackets.cfdp.pdu.helper import GenericPduPacket, PduHolder
from tmtccmd.cfdp import (
    CfdpUserBase,
    LocalEntityCfg,
    RemoteEntityCfgTable,
    RemoteEntityCfg,
)
from tmtccmd.cfdp.defs import CfdpStates, TransactionId
from tmtccmd.cfdp.handler.crc import Crc32Helper
from tmtccmd.cfdp.handler.defs import (
    FileParamsBase,
    PacketSendNotConfirmed,
    NoRemoteEntityCfgFound,
)
from tmtccmd.cfdp.user import (
    MetadataRecvParams,
    FileSegmentRecvdParams,
    TransactionFinishedParams,
)


_LOGGER = logging.getLogger(__name__)


@dataclass
class DestFileParams(FileParamsBase):
    file_name: Path

    @classmethod
    def empty(cls) -> DestFileParams:
        return cls(
            progress=0,
            segment_len=0,
            crc32=bytes(),
            file_size=0,
            file_name=Path(),
            no_file_data=False,
        )

    def reset(self):
        super().reset()
        self.file_name = Path()


class TransactionStep(enum.Enum):
    IDLE = 0
    # Metadata was received
    TRANSACTION_START = 1
    RECEIVING_FILE_DATA = 2
    SENDING_ACK_PDU = 3
    # File transfer complete. Perform checksum verification and notice of completion
    TRANSFER_COMPLETION = 4
    SENDING_FINISHED_PDU = 5


@dataclass
class DestStateWrapper:
    state: CfdpStates = CfdpStates.IDLE
    transaction: TransactionStep = TransactionStep.IDLE
    transaction_id: Optional[TransactionId] = None
    packet_ready: bool = False


@dataclass
class DestFieldWrapper:
    transaction_id: Optional[TransactionId] = None
    remote_cfg: Optional[RemoteEntityCfg] = None
    closure_requested: bool = False
    condition_code: ConditionCode = ConditionCode.NO_CONDITION_FIELD
    delivery_code: DeliveryCode = DeliveryCode.DATA_INCOMPLETE
    file_status: FileDeliveryStatus = FileDeliveryStatus.DISCARDED_DELIBERATELY
    pdu_conf: PduConfig = dataclasses.field(default_factory=lambda: PduConfig.empty())
    fp: DestFileParams = dataclasses.field(
        default_factory=lambda: DestFileParams.empty()
    )
    file_directives_dict: Dict[
        DirectiveType, List[AbstractFileDirectiveBase]
    ] = dataclasses.field(default_factory=lambda: dict())
    file_data_deque: Deque[FileDataPdu] = dataclasses.field(
        default_factory=lambda: deque()
    )

    def reset(self):
        self.transaction_id = None
        self.closure_requested = False
        self.condition_code = ConditionCode.NO_CONDITION_FIELD
        self.delivery_code = DeliveryCode.DATA_INCOMPLETE
        self.pdu_conf = PduConfig.empty()
        self.fp.reset()
        self.remote_cfg = None

    def clear_file_deque(self):
        self.file_data_deque.clear()

    def clear_file_directive_dict(self):
        self.file_directives_dict.clear()


class FsmResult:
    def __init__(self, states: DestStateWrapper, pdu_holder: PduHolder):
        self.states = states
        self.pdu_holder = pdu_holder


class DestHandler:
    def __init__(
        self,
        cfg: LocalEntityCfg,
        user: CfdpUserBase,
        remote_cfg_table: RemoteEntityCfgTable,
    ):
        self.cfg = cfg
        self.remote_cfg_table = remote_cfg_table
        self.states = DestStateWrapper()
        self.user = user
        self.pdu_holder = PduHolder(None)
        self._params = DestFieldWrapper()
        self._crc_helper: Crc32Helper = Crc32Helper(
            ChecksumType.NULL_CHECKSUM, user.vfs
        )

    def __transaction_start_metadata_pdu_to_params(self, metadata_pdu: MetadataPdu):
        self._params.reset()
        self.states.transaction = TransactionStep.TRANSACTION_START
        if metadata_pdu.pdu_header.trans_mode == TransmissionMode.UNACKNOWLEDGED:
            self.states.state = CfdpStates.BUSY_CLASS_1_NACKED
        elif metadata_pdu.pdu_header.trans_mode == TransmissionMode.ACKNOWLEDGED:
            self.states.state = CfdpStates.BUSY_CLASS_2_ACKED
        self._crc_helper.checksum_type = metadata_pdu.checksum_type
        self._closure_requested = metadata_pdu.closure_requested
        if metadata_pdu.dest_file_name is None:
            self._params.fp.no_file_data = True
        else:
            self._params.fp.file_name = Path(metadata_pdu.dest_file_name)
        self._params.fp.size_from_metadata = metadata_pdu.file_size
        self._params.pdu_conf = metadata_pdu.pdu_conf
        self._params.pdu_conf.direction = Direction.TOWARDS_SENDER
        self._params.transaction_id = TransactionId(
            source_entity_id=metadata_pdu.source_entity_id,
            transaction_seq_num=metadata_pdu.transaction_seq_num,
        )
        self.states.transaction_id = self._params.transaction_id
        self._params.remote_cfg = self.remote_cfg_table.get_cfg(
            metadata_pdu.source_entity_id
        )

    def __transaction_start_vfs_handling(self):
        try:
            if self.user.vfs.file_exists(self._params.fp.file_name):
                self.user.vfs.truncate_file(self._params.fp.file_name)
            else:
                self.user.vfs.create_file(self._params.fp.file_name)
        except PermissionError:
            self._params.file_status = FileDeliveryStatus.DISCARDED_FILESTORE_REJECTION
            fh = self.cfg.default_fault_handlers.get_fault_handler(
                ConditionCode.FILESTORE_REJECTION
            )
            if fh == FaultHandlerCode.IGNORE_ERROR:
                pass
            elif fh == FaultHandlerCode.ABANDON_TRANSACTION:
                # TODO: Implement
                pass
            elif fh == FaultHandlerCode.NOTICE_OF_SUSPENSION:
                # TODO: Implement
                pass
            elif fh == FaultHandlerCode.NOTICE_OF_CANCELLATION:
                # TODO: Implement
                pass
            self.cfg.default_fault_handlers.report_fault(
                ConditionCode.FILESTORE_REJECTION
            )

    def _start_transaction(self, metadata_pdu: MetadataPdu) -> bool:
        if self.states.state != CfdpStates.IDLE:
            return False
        self.__transaction_start_metadata_pdu_to_params(metadata_pdu)
        # I am not fully sure whether a remote configuration is strictly required for
        # a destination handler. I think to be fully standard-compliant or at least allow
        # the flexibility to be standard-compliant in the future, we should require that
        # a remote entity configuration exists for each CFDP sender.
        if self._params.remote_cfg is None:
            _LOGGER.warning(
                f"No remote configuration found for remote ID {metadata_pdu.dest_entity_id}"
            )
            raise NoRemoteEntityCfgFound(metadata_pdu.dest_entity_id)
        self.states.transaction = TransactionStep.RECEIVING_FILE_DATA
        self.__transaction_start_vfs_handling()
        msgs_to_user_list = None
        if metadata_pdu.options is not None:
            msgs_to_user_list = []
            for tlv in metadata_pdu.options:
                if tlv.tlv_type == TlvTypes.MESSAGE_TO_USER:
                    msgs_to_user_list.append(tlv)
        params = MetadataRecvParams(
            transaction_id=self._params.transaction_id,
            file_size=metadata_pdu.file_size,
            source_id=metadata_pdu.source_entity_id,
            dest_file_name=metadata_pdu.dest_file_name,
            source_file_name=metadata_pdu.source_file_name,
            msgs_to_user=msgs_to_user_list,
        )
        self.user.metadata_recv_indication(params)
        return True

    def __idle_fsm(self):
        clear_all_other_pdus = True
        for pdu_type, pdu_deque in self._params.file_directives_dict.items():
            if pdu_type == DirectiveType.METADATA_PDU:
                clear_metadata_deque = False
                for pdu_base in pdu_deque:
                    metadata_pdu = PduHolder(pdu_base).to_metadata_pdu()
                    self._start_transaction(metadata_pdu)
                    # CFDP 4.6.1.2.4: Any repeated metadata should be discarded.
                    if self.states.state != CfdpStates.IDLE:
                        clear_metadata_deque = True
                        break
                if clear_metadata_deque:
                    pdu_deque.clear()
            else:
                # For unacknowledged transfers, there is no lost metadata detection in place.
                # For now, simply discard all PDUs which arrive before the Metadata PDU.
                # TODO: This is tricky for acknowledged mode. This implementation writes to the
                #       virtual filestore directly. If lost segment detection is in place, all
                #       PDUs which arrive successfully after a missing metadata PDU would have
                #       to be counted as lost segments.
                # clear_all_other_pdus = False
                pass
        if self.states.state == CfdpStates.IDLE and clear_all_other_pdus:
            if self._params.file_directives_dict:
                for other_pdu in self._params.file_directives_dict:
                    _LOGGER.warning(
                        f"Received {other_pdu} PDU without "
                        f"first receiving metadata PDU first. Discarding it"
                    )
                self._params.file_directives_dict.clear()
            if self._params.file_data_deque:
                _LOGGER.warning(
                    f"Received {len(self._params.file_data_deque)} file data PDUs without "
                    f"first receiving metadata PDU first. Discarding them"
                )
                self._params.file_data_deque.clear()

    def __busy_naked_fsm(self):
        if self.states.transaction == TransactionStep.RECEIVING_FILE_DATA:
            self.__receiving_fd_pdus_nacknowledged()
        if self.states.transaction == TransactionStep.TRANSFER_COMPLETION:
            self._handle_transfer_completion()
        if self.states.transaction == TransactionStep.SENDING_FINISHED_PDU:
            self._prepare_finished_pdu()
            self.states.packet_ready = True

    def __receiving_fd_pdus_nacknowledged(self):
        # TODO: Sequence count check
        for file_data_pdu in self._params.file_data_deque:
            self.__handle_one_fd_pdu(file_data_pdu)
        # TODO: Support for check timer missing
        eof_pdus = self._params.file_directives_dict.get(DirectiveType.EOF_PDU)
        if eof_pdus is not None:
            for pdu in eof_pdus:
                eof_pdu = PduHolder(pdu).to_eof_pdu()
                self._handle_eof_pdu(eof_pdu)

    def __handle_one_fd_pdu(self, file_data_pdu: FileDataPdu):
        data = file_data_pdu.file_data
        offset = file_data_pdu.offset
        if self.cfg.indication_cfg.file_segment_recvd_indication_required:
            file_segment_indic_params = FileSegmentRecvdParams(
                transaction_id=self._params.transaction_id,
                length=len(file_data_pdu.file_data),
                offset=offset,
                record_cont_state=file_data_pdu.record_cont_state,
                segment_metadata=file_data_pdu.segment_metadata,
            )
            self.user.file_segment_recv_indication(file_segment_indic_params)
        try:
            self.user.vfs.write_data(self._params.fp.file_name, data, offset)
            self._params.file_status = FileDeliveryStatus.FILE_RETAINED
            # Ensure that the progress value is always incremented
            if offset + len(file_data_pdu.file_data) > self._params.fp.progress:
                self._params.fp.progress = offset + len(file_data_pdu.file_data)
        except FileNotFoundError:
            if self._params.file_status != FileDeliveryStatus.FILE_RETAINED:
                self._params.file_status = FileDeliveryStatus.DISCARDED_DELIBERATELY
        except PermissionError:
            if self._params.file_status != FileDeliveryStatus.FILE_RETAINED:
                self._params.file_status = (
                    FileDeliveryStatus.DISCARDED_FILESTORE_REJECTION
                )

    def state_machine(self) -> FsmResult:
        if self.states.state == CfdpStates.IDLE:
            self.__idle_fsm()
        if self.states.state == CfdpStates.BUSY_CLASS_1_NACKED:
            self.__busy_naked_fsm()
        return FsmResult(self.states, self.pdu_holder)

    def finish(self):
        self._params.reset()
        # Not fully sure this is the best approach, but I think this is ok for now
        self._params.clear_file_directive_dict()
        self._params.clear_file_deque()
        self._params.transaction_id = None
        self.states.state = CfdpStates.IDLE
        self.states.transaction = TransactionStep.IDLE

    def pass_packet(self, packet: GenericPduPacket):
        # TODO: Sanity checks
        if packet.pdu_type == PduType.FILE_DATA:
            self._params.file_data_deque.append(cast(FileDataPdu, packet))
        else:
            if packet.directive_type in self._params.file_directives_dict:
                self._params.file_directives_dict.get(packet.directive_type).append(
                    packet
                )
            else:
                self._params.file_directives_dict.update(
                    {packet.directive_type: [packet]}
                )

    def confirm_packet_sent_advance_fsm(self):
        """Helper method which performs both :py:meth:`confirm_packet_sent` and
        :py:meth:`advance_fsm`
        """
        self.confirm_packet_sent()
        self.advance_fsm()

    def confirm_packet_sent(self):
        """Confirm that a packet generated by the :py:meth:`operation` was sent successfully"""
        self.states.packet_ready = False

    def advance_fsm(self):
        if self.states.packet_ready:
            raise PacketSendNotConfirmed(
                f"Must send current packet {self.pdu_holder.base} first"
            )
        if self.states.transaction == TransactionStep.SENDING_FINISHED_PDU:
            self.finish()

    def _handle_transfer_completion(self):
        if self._crc_helper.checksum_type != ChecksumType.NULL_CHECKSUM:
            self._checksum_verify()
        elif (
            self._params.fp.no_file_data
            or self._crc_helper.checksum_type == ChecksumType.NULL_CHECKSUM
        ):
            self._params.condition_code = ConditionCode.NO_ERROR
        self._notice_of_completion()
        if self.states.state == CfdpStates.BUSY_CLASS_1_NACKED:
            if self._params.closure_requested:
                self.states.transaction = TransactionStep.SENDING_FINISHED_PDU
            else:
                self.finish()
        elif self.states.state == CfdpStates.BUSY_CLASS_2_ACKED:
            self.states.transaction = TransactionStep.SENDING_FINISHED_PDU

    def _handle_eof_pdu(self, eof_pdu: EofPdu):
        # TODO: Error handling
        if eof_pdu.condition_code == ConditionCode.NO_ERROR:
            self._params.fp.crc32 = eof_pdu.file_checksum
            file_size_from_eof = eof_pdu.file_size
            # CFDP 4.6.1.2.9: Declare file size error if progress exceeds file size
            if self._params.fp.progress > file_size_from_eof:
                # TODO: File Size error
                pass
            self._params.fp.file_size = file_size_from_eof
            self._params.fp.segment_len = self._params.remote_cfg.max_file_segment_len
            if self.cfg.indication_cfg.eof_recv_indication_required:
                self.user.eof_recv_indication(self._params.transaction_id)
            if self.states.transaction == TransactionStep.RECEIVING_FILE_DATA:
                if self.states.state == CfdpStates.BUSY_CLASS_1_NACKED:
                    self.states.transaction = TransactionStep.TRANSFER_COMPLETION
                elif self.states.state == CfdpStates.BUSY_CLASS_2_ACKED:
                    self.states.transaction = TransactionStep.SENDING_ACK_PDU

    def _checksum_verify(self):
        crc32 = self._crc_helper.calc_for_file(
            self._params.fp.file_name,
            self._params.fp.file_size,
            self._params.fp.segment_len,
        )
        if crc32 != self._params.fp.crc32:
            # TODO: CFDP Checksum error handling
            self._params.condition_code = ConditionCode.FILE_CHECKSUM_FAILURE
        else:
            self._params.delivery_code = DeliveryCode.DATA_COMPLETE
            self._params.condition_code = ConditionCode.NO_ERROR

    def _notice_of_completion(self):
        if self.cfg.indication_cfg.transaction_finished_indication_required:
            finished_indic_params = TransactionFinishedParams(
                transaction_id=self._params.transaction_id,
                condition_code=self._params.condition_code,
                delivery_code=self._params.delivery_code,
                file_status=self._params.file_status,
                status_report=None,
            )
            self.user.transaction_finished_indication(finished_indic_params)

    def _prepare_finished_pdu(self):
        if self.states.packet_ready:
            raise PacketSendNotConfirmed(
                f"Must send current packet {self.pdu_holder.base} first"
            )
        finished_params = FinishedParams(
            condition_code=self._params.condition_code,
            # TODO: Find out how those are used
            delivery_code=DeliveryCode.DATA_COMPLETE,
            delivery_status=FileDeliveryStatus.FILE_RETAINED,
        )
        finished_pdu = FinishedPdu(
            params=finished_params,
            # The configuration was cached when the first metadata arrived
            pdu_conf=self._params.pdu_conf,
        )
        self.pdu_holder.base = finished_pdu



"""TCP communication interface"""
import logging
import socket
import time
import enum
import threading
import select
from collections import deque
from typing import Optional, Sequence

from spacepackets.ccsds.spacepacket import parse_space_packets, PacketId

from tmtccmd.com import ComInterface, SendError
from tmtccmd.tm import TelemetryListT
from tmtccmd.com.tcpip_utils import EthAddr

_LOGGER = logging.getLogger(__name__)

TCP_RECV_WIRETAPPING_ENABLED = False
TCP_SEND_WIRETAPPING_ENABLED = False


class TcpCommunicationType(enum.Enum):
    """Parse for space packets in the TCP stream, using the space packet header."""

    SPACE_PACKETS = 0


class TcpSpacePacketsComIF(ComInterface):
    """Communication interface for TCP communication. This particular interface expects
    raw space packets to be sent via TCP and uses a list of passed packet IDs to parse for them.
    """

    def __init__(
        self,
        com_if_id: str,
        space_packet_ids: Sequence[PacketId],
        tm_polling_freqency: float,
        target_address: EthAddr,
        max_packets_stored: int = 50,
    ):
        """Initialize a communication interface to send and receive TMTC via TCP.

        :param com_if_id:
        :param space_packet_ids: Valid packet IDs for CCSDS space packets. Those will be used
            to parse for space packets inside the TCP stream.
        :param tm_polling_freqency: Polling frequency in seconds
        """
        self.com_if_id = com_if_id
        self.com_type = TcpCommunicationType.SPACE_PACKETS
        self.space_packet_ids = space_packet_ids
        self.tm_polling_frequency = tm_polling_freqency
        self.target_address = target_address
        self.max_packets_stored = max_packets_stored
        self.connected = False

        self.__tcp_socket: Optional[socket.socket] = None
        self.__last_connection_time = 0
        self.__tm_thread_kill_signal = threading.Event()
        # Separate thread to request TM packets periodically if no TCs are being sent
        self.__tcp_conn_thread: Optional[threading.Thread] = threading.Thread(
            target=self.__tcp_tm_client, daemon=True
        )
        self.__tm_queue = deque()
        self.__analysis_queue = deque()
        self.tm_packet_list = []

    @property
    def id(self) -> str:
        return self.com_if_id

    def __del__(self):
        try:
            self.close()
        except IOError:
            _LOGGER.warning("Could not close TCP communication interface!")

    def initialize(self, args: any = None) -> any:
        pass

    def open(self, args: any = None):
        self.__tm_thread_kill_signal.clear()
        try:
            self.set_up_socket()
        except IOError as e:
            _LOGGER.exception("Issues setting up the TCP socket")
            raise e
        self.set_up_tcp_thread()
        self.__tcp_conn_thread.start()

    def is_open(self) -> bool:
        return self.connected

    def set_up_socket(self):
        if self.__tcp_socket is None:
            self.__tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                self.__tcp_socket.settimeout(2.0)
                self.__tcp_socket.connect(self.target_address.to_tuple)
                self.connected = True
            except socket.timeout as e:
                _LOGGER.warning(
                    f"Could not connect to socket with address {self.target_address}: {e}"
                )
            finally:
                self.__tcp_socket.settimeout(None)

    def set_up_tcp_thread(self):
        if self.__tcp_conn_thread is None:
            self.__tcp_conn_thread = threading.Thread(
                target=self.__tcp_tm_client, daemon=True
            )

    def close(self, args: any = None) -> None:
        self.__tm_thread_kill_signal.set()
        socket_was_closed = False
        if self.__tcp_conn_thread is not None:
            if self.__tcp_conn_thread.is_alive():
                self.__tcp_conn_thread.join(self.tm_polling_frequency)
            if self.connected:
                try:
                    self.__tcp_socket.shutdown(socket.SHUT_RDWR)
                except OSError:
                    _LOGGER.warning(
                        "TCP socket endpoint was already closed or not connected"
                    )
                self.__tcp_socket.close()
                socket_was_closed = True
        if self.__tcp_socket is not None and not socket_was_closed:
            self.__tcp_socket.close()
        self.__tcp_socket = None
        self.__tcp_conn_thread = None

    def send(self, data: bytes):
        try:
            if not self.connected:
                self.set_up_socket()
            self.__tcp_socket.sendto(data, self.target_address.to_tuple)
        except BrokenPipeError as e:
            raise SendError(f"{e}", e)
        except ConnectionRefusedError or OSError as e:
            self.connected = False
            self.__tcp_socket.close()
            self.__tcp_socket = None
            raise SendError(f"TCP connection attempt failed with exception: {e}", e)

    def receive(self, poll_timeout: float = 0) -> TelemetryListT:
        self.__tm_queue_to_packet_list()
        tm_packet_list = self.tm_packet_list
        self.tm_packet_list = []
        return tm_packet_list

    def __tm_queue_to_packet_list(self):
        while self.__tm_queue:
            self.__analysis_queue.appendleft(self.__tm_queue.pop())
        # TCP is stream based, so there might be broken packets or multiple packets in one recv
        # call. We parse the space packets contained in the stream here
        if self.com_type == TcpCommunicationType.SPACE_PACKETS:
            self.tm_packet_list.extend(
                parse_space_packets(
                    analysis_queue=self.__analysis_queue,
                    packet_ids=self.space_packet_ids,
                )
            )
        else:
            while self.__analysis_queue:
                self.tm_packet_list.append(self.__analysis_queue.pop())

    def __tcp_tm_client(self):
        while True and not self.__tm_thread_kill_signal.is_set():
            if self.connected:
                try:
                    self.__receive_tm_packets()
                except ConnectionRefusedError:
                    _LOGGER.warning("TCP connection attempt failed..")
            time.sleep(self.tm_polling_frequency)

    def __receive_tm_packets(self):
        try:
            ready = select.select([self.__tcp_socket], [], [], 0)
            if ready[0]:
                bytes_recvd = self.__tcp_socket.recv(4096)
                if bytes_recvd == b"":
                    self.__close_tcp_socket()
                    _LOGGER.info("TCP server has been closed")
                    return
                else:
                    self.connected = True
                if self.__tm_queue.__len__() >= self.max_packets_stored:
                    _LOGGER.warning(
                        "Number of packets in TCP queue too large. "
                        "Overwriting old packets.."
                    )
                    self.__tm_queue.pop()
                    # TODO: If segments are received but the receiver is unable to parse packets
                    #       properly, it might make sense to have a timeout which then also
                    #       logs that there might be an issue reading packets
                self.__tm_queue.appendleft(bytes(bytes_recvd))
        except ConnectionResetError:
            self.__close_tcp_socket()
            _LOGGER.exception("ConnectionResetError. TCP server might not be up")

    def data_available(self, timeout: float = 0, parameters: any = 0) -> int:
        self.__tm_queue_to_packet_list()
        return len(self.tm_packet_list)

    def __close_tcp_socket(self):
        self.connected = False
        self.__tcp_socket.close()
        self.__tcp_socket = None



#!/usr/bin/env python3
"""Example application for the TMTC Commander"""
import logging
import sys
import time
from typing import Optional

import tmtccmd
from spacepackets.ccsds import CdsShortTimestamp
from spacepackets.ecss import PusTelemetry, PusTelecommand, PusVerificator
from spacepackets.ecss.pus_17_test import Service17Tm
from spacepackets.ecss.pus_1_verification import UnpackParams, Service1Tm
from spacepackets.util import UnsignedByteField

from tmtccmd import CcsdsTmtcBackend, ProcedureParamsWrapper, BackendRequest
from tmtccmd.com import ComInterface
from tmtccmd.pus import VerificationWrapper
from tmtccmd.tm import CcsdsTmHandler, SpecificApidHandlerBase
from tmtccmd.config import (
    default_json_path,
    SetupParams,
    HookBase,
    TmtcDefinitionWrapper,
    CoreServiceList,
    OpCodeEntry,
    params_to_procedure_conversion,
    PreArgsParsingWrapper,
    SetupWrapper,
)
from tmtccmd.logging import add_colorlog_console_logger
from tmtccmd.logging.pus import (
    RegularTmtcLogWrapper,
    RawTmtcTimedLogWrapper,
    TimedLogWhen,
)
from tmtccmd.tc import (
    TcQueueEntryType,
    ProcedureWrapper,
    TcProcedureType,
    FeedWrapper,
    SendCbParams,
    DefaultPusQueueHelper,
    TcHandlerBase,
    QueueWrapper,
)
from tmtccmd.tm.pus_5_fsfw_event import Service5Tm
from tmtccmd.util import FileSeqCountProvider, PusFileSeqCountProvider, ObjectIdDictT
from tmtccmd.util.tmtc_printer import FsfwTmTcPrinter

_LOGGER = logging.getLogger()

EXAMPLE_PUS_APID = 0xEF
EXAMPLE_CFDP_APID = 0xF0

CFDP_LOCAL_ENTITY_ID = UnsignedByteField(byte_len=2, val=1)
CFDP_REMOTE_ENTITY_ID = UnsignedByteField(byte_len=2, val=EXAMPLE_CFDP_APID)


class ExampleHookClass(HookBase):
    def __init__(self, json_cfg_path: str):
        super().__init__(json_cfg_path=json_cfg_path)

    def assign_communication_interface(self, com_if_key: str) -> Optional[ComInterface]:
        print("Communication interface assignment function was called")
        from tmtccmd.config.com import (
            create_com_interface_default,
            create_com_interface_cfg_default,
        )

        cfg = create_com_interface_cfg_default(
            com_if_key=com_if_key,
            json_cfg_path=self.cfg_path,
            space_packet_ids=None,
        )
        return create_com_interface_default(cfg)

    def get_tmtc_definitions(self) -> TmtcDefinitionWrapper:
        from tmtccmd.config.globals import get_default_tmtc_defs

        defs = get_default_tmtc_defs()
        srv_5 = OpCodeEntry()
        srv_5.add("0", "Event Test")
        defs.add_service(
            name=CoreServiceList.SERVICE_5.value,
            info="PUS Service 5 Event",
            op_code_entry=srv_5,
        )
        srv_17 = OpCodeEntry()
        srv_17.add("0", "Ping Test")
        defs.add_service(
            name=CoreServiceList.SERVICE_17_ALT,
            info="PUS Service 17 Test",
            op_code_entry=srv_17,
        )
        return defs

    def perform_mode_operation(self, tmtc_backend: CcsdsTmtcBackend, mode: int):
        _LOGGER.info("Mode operation hook was called")
        pass

    def get_object_ids(self) -> ObjectIdDictT:
        from tmtccmd.config.objects import get_core_object_ids

        return get_core_object_ids()


class DualLogger:
    def __init__(self, file_logger: logging.Logger, console_logger: logging.Logger):
        self.file_logger = file_logger
        self.console_logger = console_logger

    def info(self, string: str):
        self.console_logger.info(string)
        self.file_logger.info(string)

    def print(self, string: str):
        print(string)
        self.file_logger.info(string)


class PusTmHandler(SpecificApidHandlerBase):
    def __init__(
        self,
        verif_wrapper: VerificationWrapper,
        printer: FsfwTmTcPrinter,
        raw_logger: RawTmtcTimedLogWrapper,
    ):
        super().__init__(EXAMPLE_PUS_APID, None)
        self.printer = printer
        self.raw_logger = raw_logger
        self.verif_wrapper = verif_wrapper
        self.dlogger = DualLogger(self.printer.file_logger, _LOGGER)

    def handle_tm(self, packet: bytes, _user_args: any):
        try:
            tm_packet = PusTelemetry.unpack(
                packet, time_reader=CdsShortTimestamp.empty()
            )
        except ValueError as e:
            _LOGGER.warning("Could not generate PUS TM object from raw data")
            _LOGGER.warning(f"Raw Packet: [{packet.hex(sep=',')}], REPR: {packet!r}")
            raise e
        service = tm_packet.service
        dedicated_handler = False
        if service == 1:
            verif_tm = Service1Tm.unpack(
                data=packet, params=UnpackParams(CdsShortTimestamp.empty(), 1, 2)
            )
            res = self.verif_wrapper.add_tm(verif_tm)
            if res is None:
                _LOGGER.info(
                    f"Received Verification TM[{tm_packet.service}, {tm_packet.subservice}] "
                    f"with Request ID {verif_tm.tc_req_id.as_u32():#08x}"
                )
                _LOGGER.warning(
                    f"No matching telecommand found for {verif_tm.tc_req_id}"
                )
            else:
                self.verif_wrapper.log_to_console(verif_tm, res)
                self.verif_wrapper.log_to_file(verif_tm, res)
            dedicated_handler = True
        if service == 5:
            event_tm = Service5Tm.unpack(packet, time_reader=CdsShortTimestamp.empty())
            self.dlogger.info(
                f"Received event packet TM [{event_tm.service}, {event_tm.subservice}]"
            )
            dedicated_handler = True
        if service == 17:
            ping_tm = Service17Tm.unpack(packet, time_reader=CdsShortTimestamp.empty())
            dedicated_handler = True
            if ping_tm.subservice == 2:
                self.dlogger.info("Received Ping Reply TM[17,2]")
            else:
                self.dlogger.info(
                    f"Received Test Packet with unknown subservice {tm_packet.subservice}"
                )
        if tm_packet is None:
            _LOGGER.info(
                f"The service {service} is not implemented in Telemetry Factory"
            )
            tm_packet = PusTelemetry.unpack(
                packet, time_reader=CdsShortTimestamp.empty()
            )
        self.raw_logger.log_tm(tm_packet)
        if not dedicated_handler and tm_packet is not None:
            pass
            # TODO: Create more generic print method..
            # self.printer.handle_long_tm_print(packet_if=tm_packet, info_if=tm_packet)


class TcHandler(TcHandlerBase):
    def __init__(
        self,
        seq_count_provider: FileSeqCountProvider,
        verif_wrapper: VerificationWrapper,
    ):
        super(TcHandler, self).__init__()
        self.seq_count_provider = seq_count_provider
        self.verif_wrapper = verif_wrapper
        self.queue_helper = DefaultPusQueueHelper(
            queue_wrapper=QueueWrapper.empty(),
            seq_cnt_provider=seq_count_provider,
            pus_verificator=self.verif_wrapper.pus_verificator,
            tc_sched_timestamp_len=7,
            default_pus_apid=EXAMPLE_PUS_APID,
        )

    def send_cb(self, send_params: SendCbParams):
        entry_helper = send_params.entry
        if entry_helper.is_tc:
            if entry_helper.entry_type == TcQueueEntryType.PUS_TC:
                pus_tc_wrapper = entry_helper.to_pus_tc_entry()
                pus_tc_wrapper.pus_tc.seq_count = (
                    self.seq_count_provider.get_and_increment()
                )
                self.verif_wrapper.add_tc(pus_tc_wrapper.pus_tc)
                raw_tc = pus_tc_wrapper.pus_tc.pack()
                _LOGGER.info(f"Sending {pus_tc_wrapper.pus_tc}")
                send_params.com_if.send(raw_tc)
        elif entry_helper.entry_type == TcQueueEntryType.LOG:
            log_entry = entry_helper.to_log_entry()
            _LOGGER.info(log_entry.log_str)

    def queue_finished_cb(self, helper: ProcedureWrapper):
        if helper.proc_type == TcProcedureType.DEFAULT:
            def_proc = helper.to_def_procedure()
            _LOGGER.info(
                f"Queue handling finished for service {def_proc.service} and "
                f"op code {def_proc.op_code}"
            )

    def feed_cb(self, helper: ProcedureWrapper, wrapper: FeedWrapper):
        self.queue_helper.queue_wrapper = wrapper.queue_wrapper
        if helper.proc_type == TcProcedureType.DEFAULT:
            def_proc = helper.to_def_procedure()
            service = def_proc.service
            if (
                service == CoreServiceList.SERVICE_17
                or service == CoreServiceList.SERVICE_17_ALT
            ):
                return self.queue_helper.add_pus_tc(
                    PusTelecommand(service=17, subservice=1)
                )


# Note about lint disable: I could split up the function but I prefer to have the whole
# running main in one block for the example.
def main():  # noqa: C901
    # Apply the library logger formatting to the own root logger. Library logs will be propagated
    # to application logger.
    add_colorlog_console_logger(_LOGGER)
    tmtccmd.init_printout(False)
    hook_obj = ExampleHookClass(json_cfg_path=default_json_path())
    parser_wrapper = PreArgsParsingWrapper()
    parser_wrapper.create_default_parent_parser()
    parser_wrapper.create_default_parser()
    parser_wrapper.add_def_proc_args()
    params = SetupParams()
    post_args_wrapper = parser_wrapper.parse(hook_obj=hook_obj, setup_params=params)
    proc_wrapper = ProcedureParamsWrapper()
    if post_args_wrapper.use_gui:
        post_args_wrapper.set_params_without_prompts(proc_wrapper)
    else:
        post_args_wrapper.set_params_with_prompts(proc_wrapper)
    params.apid = EXAMPLE_PUS_APID
    setup_args = SetupWrapper(
        hook_obj=hook_obj, setup_params=params, proc_param_wrapper=proc_wrapper
    )
    # Create console logger helper and file loggers
    tmtc_logger = RegularTmtcLogWrapper()
    printer = FsfwTmTcPrinter(tmtc_logger.logger)
    raw_logger = RawTmtcTimedLogWrapper(when=TimedLogWhen.PER_HOUR, interval=1)
    verificator = PusVerificator()
    verification_wrapper = VerificationWrapper(
        verificator, _LOGGER, printer.file_logger
    )
    # Create primary TM handler and add it to the CCSDS Packet Handler
    tm_handler = PusTmHandler(verification_wrapper, printer, raw_logger)
    ccsds_handler = CcsdsTmHandler(generic_handler=None)
    ccsds_handler.add_apid_handler(tm_handler)

    # Create TC handler
    seq_count_provider = PusFileSeqCountProvider()
    tc_handler = TcHandler(seq_count_provider, verification_wrapper)
    tmtccmd.setup(setup_args=setup_args)
    init_proc = params_to_procedure_conversion(setup_args.proc_param_wrapper)
    tmtc_backend = tmtccmd.create_default_tmtc_backend(
        setup_wrapper=setup_args,
        tm_handler=ccsds_handler,
        tc_handler=tc_handler,
        init_procedure=init_proc,
    )
    tmtccmd.start(tmtc_backend=tmtc_backend, hook_obj=hook_obj)
    try:
        while True:
            state = tmtc_backend.periodic_op(None)
            if state.request == BackendRequest.TERMINATION_NO_ERROR:
                sys.exit(0)
            elif state.request == BackendRequest.DELAY_IDLE:
                _LOGGER.info("TMTC Client in IDLE mode")
                time.sleep(3.0)
            elif state.request == BackendRequest.DELAY_LISTENER:
                time.sleep(0.8)
            elif state.request == BackendRequest.DELAY_CUSTOM:
                if state.next_delay.total_seconds() <= 0.4:
                    time.sleep(state.next_delay.total_seconds())
                else:
                    time.sleep(0.4)
            elif state.request == BackendRequest.CALL_NEXT:
                pass
    except KeyboardInterrupt:
        sys.exit(0)


if __name__ == "__main__":
    main()



from spacepackets.cfdp import ConditionCode
from tmtccmd.cfdp import CfdpUserBase, TransactionId
from tmtccmd.cfdp.user import (
    FileSegmentRecvdParams,
    MetadataRecvParams,
    TransactionFinishedParams,
)


class CfdpUser(CfdpUserBase):
    def __init__(self):
        super().__init__()

    def transaction_indication(self, transaction_id: TransactionId):
        pass

    def transaction_finished_indication(self, params: TransactionFinishedParams):
        pass

    def eof_sent_indication(self, transaction_id: TransactionId):
        pass

    def abandon_indication(
        self, transaction_id: int, cond_code: ConditionCode, progress: int
    ):
        pass

    def metadata_recv_indication(self, params: MetadataRecvParams):
        pass

    def file_segment_recv_indication(self, params: FileSegmentRecvdParams):
        pass

    def report_indication(self, transaction_id: TransactionId, status_report: any):
        pass

    def suspended_indication(
        self, transaction_id: TransactionId, cond_code: ConditionCode
    ):
        pass

    def resumed_indication(self, transaction_id: TransactionId, progress: int):
        pass

    def fault_indication(
        self, transaction_id: TransactionId, cond_code: ConditionCode, progress: int
    ):
        pass

    def abandoned_indication(
        self, transaction_id: TransactionId, cond_code: ConditionCode, progress: int
    ):
        pass

    def eof_recv_indication(self, transaction_id: TransactionId):
        pass



import os
import random
import sys

from crcmod.predefined import PredefinedCrc

from spacepackets.util import ByteFieldU16, ByteFieldU8
from tmtccmd.cfdp.defs import CfdpStates
from tmtccmd.cfdp.handler import SourceHandler, FsmResult
from tmtccmd.cfdp.handler.source import TransactionStep
from tmtccmd.cfdp.request import PutRequest, PutRequestCfg
from .test_src_handler import TestCfdpSourceHandler


class TestCfdpSourceHandlerNoClosure(TestCfdpSourceHandler):
    def setUp(self) -> None:
        self.common_setup(False)

    def test_empty_file(self):
        self._common_empty_file_test()
        fsm_res = self.source_handler.state_machine()
        self.cfdp_user.transaction_finished_indication.assert_called_once()
        self.assertEqual(self.cfdp_user.transaction_finished_indication.call_count, 1)
        self.source_handler.confirm_packet_sent_advance_fsm()
        self.assertEqual(fsm_res.states.state, CfdpStates.IDLE)
        self.assertEqual(fsm_res.states.step, TransactionStep.IDLE)

    def test_small_file_pdu_generation(self):
        file_content = "Hello World\n"
        self._common_small_file_test(False, file_content)
        self._verify_eof_indication()
        self._test_transaction_completion()

    def test_perfectly_segmented_file_pdu_generation(self):
        # This tests generates two file data PDUs
        if sys.version_info >= (3, 9):
            rand_data = random.randbytes(self.file_segment_len * 2)
        else:
            rand_data = os.urandom(self.file_segment_len * 2)
        self.source_id = ByteFieldU8(1)
        self.dest_id = ByteFieldU8(2)
        self.source_handler.source_id = self.source_id
        dest_path = "/tmp/hello_two_segments_copy.txt"
        file_size, crc32 = self._transaction_with_file_data_wrapper(
            dest_path, rand_data
        )
        self._first_file_segment_handling(self.source_handler, rand_data)
        file_data_pdu = self._second_file_segment_handling(self.source_handler)
        self.assertEqual(len(file_data_pdu.file_data), self.file_segment_len)
        self.assertEqual(
            file_data_pdu.file_data[0 : self.file_segment_len],
            rand_data[self.file_segment_len :],
        )
        self.assertEqual(file_data_pdu.offset, self.file_segment_len)
        self.source_handler.confirm_packet_sent_advance_fsm()
        fsm_res = self.source_handler.state_machine()
        self._test_eof_file_pdu(fsm_res, file_size, crc32)
        self.source_handler.confirm_packet_sent_advance_fsm()
        self._test_transaction_completion()

    def test_segmented_file_pdu_generation(self):
        # This tests generates two file data PDUs, but the second one does not have a
        # full segment length
        if sys.version_info >= (3, 9):
            rand_data = random.randbytes(round(self.file_segment_len * 1.3))
        else:
            rand_data = os.urandom(round(self.file_segment_len * 1.3))
        remainder_len = len(rand_data) - self.file_segment_len
        self.source_id = ByteFieldU16(1)
        self.dest_id = ByteFieldU16(2)
        self.source_handler.source_id = self.source_id
        dest_path = "/tmp/hello_two_segments_imperfect_copy.txt"
        file_size, crc32 = self._transaction_with_file_data_wrapper(
            dest_path, rand_data
        )
        self._first_file_segment_handling(self.source_handler, rand_data)
        file_data_pdu = self._second_file_segment_handling(self.source_handler)
        self.assertEqual(len(file_data_pdu.file_data), remainder_len)
        self.assertEqual(
            file_data_pdu.file_data,
            rand_data[self.file_segment_len :],
        )
        self.assertEqual(file_data_pdu.offset, self.file_segment_len)
        self.source_handler.confirm_packet_sent_advance_fsm()
        fsm_res = self.source_handler.state_machine()
        self._test_eof_file_pdu(fsm_res, file_size, crc32)
        self.source_handler.confirm_packet_sent_advance_fsm()
        self._test_transaction_completion()

    def _second_file_segment_handling(self, source_handler: SourceHandler):
        source_handler.confirm_packet_sent_advance_fsm()
        fsm_res = source_handler.state_machine()
        self.assertEqual(fsm_res.states.state, CfdpStates.BUSY_CLASS_1_NACKED)
        self.assertEqual(fsm_res.states.step, TransactionStep.SENDING_FILE_DATA)
        self.assertFalse(fsm_res.pdu_holder.is_file_directive)
        return fsm_res.pdu_holder.to_file_data_pdu()

    def _transaction_with_file_data_wrapper(
        self, dest_path: str, data: bytes
    ) -> (int, bytes):
        put_req_cfg = PutRequestCfg(
            destination_id=self.dest_id,
            source_file=self.file_path,
            dest_file=dest_path,
            # Let the transmission mode be auto-determined by the remote MIB
            trans_mode=None,
            closure_requested=False,
        )
        with open(self.file_path, "wb") as of:
            crc32 = PredefinedCrc("crc32")
            crc32.update(data)
            crc32 = crc32.digest()
            of.write(data)
        file_size = self.file_path.stat().st_size
        self.local_cfg.local_entity_id = self.source_id
        self._start_source_transaction(self.dest_id, PutRequest(put_req_cfg))
        return file_size, crc32

    def _first_file_segment_handling(
        self, source_handler: SourceHandler, all_data: bytes
    ):
        fsm_res = source_handler.state_machine()
        file_data_pdu = self._check_fsm_and_contained_file_data(fsm_res)
        self.assertEqual(len(file_data_pdu.file_data), self.file_segment_len)
        self.assertEqual(
            file_data_pdu.file_data[0 : self.file_segment_len],
            all_data[0 : self.file_segment_len],
        )
        self.assertEqual(file_data_pdu.offset, 0)

    def _test_eof_file_pdu(self, fsm_res: FsmResult, file_size: int, crc32: bytes):
        self.assertEqual(fsm_res.states.state, CfdpStates.BUSY_CLASS_1_NACKED)
        self.assertEqual(fsm_res.states.step, TransactionStep.SENDING_EOF)
        eof_pdu = fsm_res.pdu_holder.to_eof_pdu()
        self.assertEqual(eof_pdu.file_size, file_size)
        self.assertEqual(eof_pdu.file_checksum, crc32)

    def _test_transaction_completion(self):
        fsm_res = self.source_handler.state_machine()
        self.assertEqual(fsm_res.states.state, CfdpStates.IDLE)
        self.assertEqual(fsm_res.states.step, TransactionStep.IDLE)
        self.cfdp_user.transaction_finished_indication.assert_called_once()
        self.assertEqual(self.cfdp_user.transaction_finished_indication.call_count, 1)



import abc
import enum
from abc import ABC
from dataclasses import dataclass
from typing import Optional, Dict, Sequence

from spacepackets.cfdp.defs import (
    FaultHandlerCode,
    ChecksumType,
    TransmissionMode,
    CFDP_VERSION_2,
    ConditionCode,
)
from spacepackets.util import UnsignedByteField
from tmtccmd.util.countdown import Countdown


class DefaultFaultHandlerBase(ABC):
    """This base class provides a way to implement the fault handling procedures as specified
    in chapter 4.8 of the CFDP standard.

    It is passed into the CFDP handlers as part of the local entity configuration and provides
    a way to specify custom user error handlers.

    It does so by mapping each applicable CFDP :py:class:`ConditionCode` to a fault handler which
    is denoted by the four :py:class:`FaultHandlerCodes`. This code is used to dispatch
    to a user-provided callback function:

     1. `FaultHandlerCodes.IGNORE_ERROR` -> :py:meth:`ignore_cb`
     2. `FaultHandlerCodes.NOTICE_OF_CANCELLATION` -> :py:meth:`notice_of_cancellation_cb`
     3. `FaultHandlerCodes.NOTICE_OF_SUSPENSION` -> :py:meth:`notice_of_suspension_cb`
     4. `FaultHandlerCodes.ABANDON_TRANSACTION` -> :py:meth:`abandon_transaction_cb`

    For each error reported by :py:meth:`report_error`, the appropriate fault handler callback
    will be called. The user provides the callbacks by providing a custom class which implements
    these base class and all abstract fault handler callbacks.
    """

    def __init__(self):
        # The initial default handle will be to ignore the error
        self._handler_dict: Dict[ConditionCode, FaultHandlerCode] = {
            ConditionCode.POSITIVE_ACK_LIMIT_REACHED: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.KEEP_ALIVE_LIMIT_REACHED: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.INVALID_TRANSMISSION_MODE: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.FILE_CHECKSUM_FAILURE: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.FILE_SIZE_ERROR: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.FILESTORE_REJECTION: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.NAK_LIMIT_REACHED: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.INACTIVITY_DETECTED: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.CHECK_LIMIT_REACHED: FaultHandlerCode.IGNORE_ERROR,
            ConditionCode.UNSUPPORTED_CHECKSUM_TYPE: FaultHandlerCode.IGNORE_ERROR,
        }

    def get_fault_handler(self, condition: ConditionCode) -> Optional[FaultHandlerCode]:
        return self._handler_dict.get(condition)

    def set_handler(self, condition: ConditionCode, handler: FaultHandlerCode):
        if condition not in self._handler_dict:
            return
        self._handler_dict.update({condition: handler})

    def report_fault(self, condition: ConditionCode):
        if condition not in self._handler_dict:
            return
        fh_code = self._handler_dict.get(condition)
        if fh_code == FaultHandlerCode.NOTICE_OF_CANCELLATION:
            self.notice_of_cancellation_cb(condition)
        elif fh_code == FaultHandlerCode.NOTICE_OF_SUSPENSION:
            self.notice_of_suspension_cb(condition)
        elif fh_code == FaultHandlerCode.IGNORE_ERROR:
            self.ignore_cb(condition)
        elif fh_code == FaultHandlerCode.ABANDON_TRANSACTION:
            self.abandoned_cb(condition)

    @abc.abstractmethod
    def notice_of_suspension_cb(self, cond: ConditionCode):
        pass

    @abc.abstractmethod
    def notice_of_cancellation_cb(self, cond: ConditionCode):
        pass

    @abc.abstractmethod
    def abandoned_cb(self, cond: ConditionCode):
        pass

    @abc.abstractmethod
    def ignore_cb(self, cond: ConditionCode):
        pass


class EntityType(enum.IntEnum):
    SENDING = 0
    RECEIVING = 1


class CheckLimitProvider(ABC):
    @abc.abstractmethod
    def provide_check_limit(
        self,
        local_entity_id: UnsignedByteField,
        remote_entity_id: UnsignedByteField,
        entity_type: EntityType,
    ) -> Countdown:
        pass


@dataclass
class IndicationCfg:
    eof_sent_indication_required: bool = True
    eof_recv_indication_required: bool = True
    file_segment_recvd_indication_required: bool = True
    transaction_finished_indication_required: bool = True
    suspended_indication_required: bool = True
    resumed_indication_required: bool = True


@dataclass
class LocalEntityCfg:
    local_entity_id: UnsignedByteField
    indication_cfg: IndicationCfg
    default_fault_handlers: DefaultFaultHandlerBase


@dataclass
class RemoteEntityCfg:
    entity_id: UnsignedByteField
    max_file_segment_len: int
    closure_requested: bool
    crc_on_transmission: bool
    default_transmission_mode: TransmissionMode
    crc_type: ChecksumType
    check_limit: Optional[CheckLimitProvider]
    # NOTE: Only this version is supported
    cfdp_version: int = CFDP_VERSION_2


class RemoteEntityCfgTable:
    def __init__(self):
        self._remote_entity_dict = dict()

    def add_config(self, cfg: RemoteEntityCfg) -> bool:
        if cfg.entity_id in self._remote_entity_dict:
            return False
        self._remote_entity_dict.update({cfg.entity_id: cfg})
        return True

    def add_configs(self, cfgs: Sequence[RemoteEntityCfg]):
        for cfg in cfgs:
            if cfg.entity_id in self._remote_entity_dict:
                continue
            self._remote_entity_dict.update({cfg.entity_id: cfg})

    def get_cfg(self, remote_entity_id: UnsignedByteField) -> RemoteEntityCfg:
        return self._remote_entity_dict.get(remote_entity_id)



"""`FSFW <https://egit.irs.uni-stuttgart.de/fsfw/fsfw>`_ specific PUS actions support"""
from .s8_fsfw_funccmd_defs import *  # noqa re-export
from tmtccmd.tc.pus_8_fsfw_funccmd import *  # noqa re-export
from tmtccmd.tm.pus_8_fsfw_funccmd import *  # noqa re-export



from unittest.mock import MagicMock

from spacepackets.cfdp import ConditionCode, Direction
from spacepackets.cfdp.pdu import FinishedPdu, FileDeliveryStatus, DeliveryCode
from spacepackets.cfdp.pdu.finished import FinishedParams
from spacepackets.util import UnsignedByteField, ByteFieldU16, ByteFieldEmpty
from tmtccmd.cfdp.defs import CfdpStates
from tmtccmd.cfdp.handler.defs import (
    InvalidPduDirection,
    InvalidSourceId,
    InvalidDestinationId,
)
from tmtccmd.cfdp.handler.source import TransactionStep
from tmtccmd.cfdp.request import PutRequestCfg, PutRequest
from .test_src_handler import TestCfdpSourceHandler


class TestCfdpSourceHandlerWithClosure(TestCfdpSourceHandler):
    def setUp(self) -> None:
        self.common_setup(True)
        self.seq_num_provider.get_and_increment = MagicMock(return_value=2)

    def test_empty_file_pdu_generation(self):
        self._common_empty_file_test()
        self._pass_simple_finish_pdu_to_source_handler()
        # Transaction should be finished
        fsm_res = self.source_handler.state_machine()
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)

    def test_small_file_pdu_generation(self):
        file_content = "Hello World\n"
        self._common_small_file_test(True, file_content)
        self._verify_eof_indication()
        self.source_handler.state_machine()
        self._pass_simple_finish_pdu_to_source_handler()
        # Transaction should be finished
        fsm_res = self.source_handler.state_machine()
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)

    def test_invalid_dir_pdu_passed(self):
        dest_id = ByteFieldU16(2)
        self._start_source_transaction(dest_id, self._prepare_dummy_put_req(dest_id))
        finish_pdu = self._prepare_finish_pdu()
        finish_pdu.pdu_file_directive.pdu_header.direction = Direction.TOWARDS_RECEIVER
        with self.assertRaises(InvalidPduDirection):
            self.source_handler.pass_packet(finish_pdu)

    def test_invalid_source_id_pdu_passed(self):
        dest_id = ByteFieldU16(2)
        finish_pdu = self._regular_transaction_start(dest_id)
        finish_pdu.pdu_file_directive.pdu_conf.source_entity_id = ByteFieldEmpty()
        with self.assertRaises(InvalidSourceId) as cm:
            self.source_handler.pass_packet(finish_pdu)
        exception = cm.exception
        self.assertEqual(exception.found_src_id, ByteFieldEmpty())
        self.assertEqual(exception.expected_src_id, ByteFieldU16(1))

    def test_invalid_dest_id_pdu_passed(self):
        dest_id = ByteFieldU16(3)
        finish_pdu = self._regular_transaction_start(dest_id)
        finish_pdu.pdu_file_directive.pdu_conf.dest_entity_id = ByteFieldEmpty()
        with self.assertRaises(InvalidDestinationId) as cm:
            self.source_handler.pass_packet(finish_pdu)
        exception = cm.exception
        self.assertEqual(exception.found_dest_id, ByteFieldEmpty())
        self.assertEqual(exception.expected_dest_id, ByteFieldU16(3))

    def _regular_transaction_start(self, dest_id: UnsignedByteField) -> FinishedPdu:
        self._start_source_transaction(dest_id, self._prepare_dummy_put_req(dest_id))
        finish_pdu = self._prepare_finish_pdu()
        self.assertEqual(finish_pdu.transaction_seq_num.value, 2)
        return finish_pdu

    def _prepare_dummy_put_req(self, dest_id: UnsignedByteField) -> PutRequest:
        return PutRequest(
            PutRequestCfg(
                destination_id=dest_id,
                source_file=self.file_path,
                dest_file="dummy.txt",
                # Let the transmission mode be auto-determined by the remote MIB
                trans_mode=None,
                closure_requested=None,
            )
        )

    def _pass_simple_finish_pdu_to_source_handler(self):
        self._state_checker(
            None, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.WAIT_FOR_FINISH
        )
        self.source_handler.pass_packet(self._prepare_finish_pdu())

    def _prepare_finish_pdu(self):
        reply_conf = self.source_handler.pdu_conf
        reply_conf.direction = Direction.TOWARDS_SENDER
        # reply_conf.dest_entity_id = dest_id
        params = FinishedParams(
            delivery_code=DeliveryCode.DATA_COMPLETE,
            condition_code=ConditionCode.NO_ERROR,
            delivery_status=FileDeliveryStatus.FILE_RETAINED,
        )
        return FinishedPdu(
            params=params,
            pdu_conf=reply_conf,
        )



import logging
import time

import deprecation

from tmtccmd import __version__
from tmtccmd.com import ComInterface
from tmtccmd.com.serial_base import SerialComBase, SerialCfg
from tmtccmd.tm import TelemetryListT


# TODO: This should be configurable
SERIAL_FRAME_MAX_LENGTH = 4096


# TODO: Consider removing this? Sending binary data without some sort of transport layer
#       is not really the best approach..
@deprecation.deprecated(
    deprecated_in="4.0.0a0",
    current_version=__version__,
    details="Please use a serial interface using a transport layer like SerialCobsComIF",
)
class SerialFixedFrameComIF(SerialComBase, ComInterface):
    def __init__(self, ser_cfg: SerialCfg):
        super().__init__(logging.getLogger(__name__), ser_cfg=ser_cfg)
        # Set to default value.
        self.serial_frame_size = SERIAL_FRAME_MAX_LENGTH

    def set_fixed_frame_settings(self, serial_frame_size: int):
        self.serial_frame_size = serial_frame_size

    @property
    def id(self) -> str:
        return self.ser_cfg.com_if_id

    def initialize(self, args: any = None) -> any:
        pass

    def open(self, args: any = None) -> None:
        super().open_port()

    def is_open(self) -> bool:
        return super().is_port_open()

    def close(self, args: any = None) -> None:
        return super().close_port()

    def send(self, data: bytes):
        self.serial.write(data)

    def receive(self, parameters: any = 0) -> TelemetryListT:
        packet_list = []
        if self.data_available(0, None):
            data = self.serial.read(self.serial_frame_size)
            pus_data_list = poll_pus_packets_fixed_frames(bytearray(data))
            for pus_packet in pus_data_list:
                packet_list.append(pus_packet)
        return packet_list

    def data_available(self, timeout: float, parameters: any) -> int:
        sleep_time = timeout / 3.0
        return self.data_available_fixed_frame(timeout=timeout, sleep_time=sleep_time)

    def data_available_fixed_frame(self, timeout: float, sleep_time: float):
        if timeout > 0:
            start_time = time.time()
            elapsed_time = 0
            while elapsed_time < timeout:
                if self.serial.inWaiting() > 0:
                    return self.serial.inWaiting()
                elapsed_time = time.time() - start_time
                time.sleep(sleep_time)
        if self.serial.inWaiting() > 0:
            return self.serial.inWaiting()


def poll_pus_packets_fixed_frames(data: bytearray) -> list:
    pus_data_list = []
    if len(data) == 0:
        return pus_data_list

    payload_length = data[4] << 8 | data[5]
    packet_size = payload_length + 7
    if payload_length == 0:
        return []
    read_size = len(data)
    pus_data = data[0:packet_size]
    pus_data_list.append(pus_data)

    SerialFixedFrameComIF.read_multiple_packets(
        data, packet_size, read_size, pus_data_list
    )
    return pus_data_list


def read_multiple_packets(
    data: bytearray, start_index: int, frame_size: int, pus_data_list: list
):
    while start_index < frame_size:
        start_index = SerialFixedFrameComIF.parse_next_packets(
            data, start_index, frame_size, pus_data_list
        )


def parse_next_packets(
    data: bytearray, start_index: int, frame_size: int, pus_data_list: list
) -> int:
    next_payload_len = data[start_index + 4] << 8 | data[start_index + 5]
    if next_payload_len == 0:
        end_index = frame_size
        return end_index
    next_packet_size = next_payload_len + 7

    if next_packet_size > SERIAL_FRAME_MAX_LENGTH:
        logger = logging.getLogger(__name__)
        logger.error(
            "PUS Polling: Very large packet detected, packet splitting not implemented yet!"
        )
        logger.error("Detected Size: " + str(next_packet_size))
        end_index = frame_size
        return end_index

    end_index = start_index + next_packet_size
    pus_data = data[start_index:end_index]
    pus_data_list.append(pus_data)
    return end_index



"""Contains classes and functions to handle PUS Service 8 telemetry.
"""
from __future__ import annotations
import struct
from typing import Optional

from spacepackets.ccsds.time import CcsdsTimeProvider
from spacepackets.ecss import PusService
from spacepackets.ecss.tm import PusTelemetry
from spacepackets.util import UnsignedByteField
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.util.obj_id import ObjectIdU32


class Service8FsfwTm(PusTmBase, PusTmInfoBase):
    """Custom Action Service Telemetry handler tailored towards Flight Software Framework (FSFW)
    TM service 8 packets
    """

    def __init__(
        self,
        subservice: int,
        object_id: bytearray,
        action_id: int,
        custom_data: bytearray,
        time: Optional[CcsdsTimeProvider] = None,
        ssc: int = 0,
        apid: int = -1,
        packet_version: int = 0b000,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        """This class can be used to deserialize service 8 packets.
        :raises ValueError: If the length of the passed bytearray is too short.
        """
        self._object_id = ObjectIdU32.from_bytes(obj_id_as_bytes=object_id)
        self._action_id = action_id
        self._custom_data = custom_data
        source_data = bytearray()
        source_data.extend(object_id)
        source_data.extend(struct.pack("!I", self._action_id))
        source_data.extend(self._custom_data)
        pus_tm = PusTelemetry(
            service=PusService.S8_FUNC_CMD,
            subservice=subservice,
            time_provider=time,
            seq_count=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.__init_without_base(instance=self)

    @staticmethod
    def __init_without_base(instance: Service8FsfwTm):
        if instance.subservice == 130:
            tm_data = instance.tm_data
            if len(tm_data) < 8:
                raise ValueError(
                    f"Length of Service 8 TM data field {len(tm_data)} short than 8"
                )
            instance.set_packet_info("Functional Data Reply")
            instance._object_id = ObjectIdU32.from_bytes(obj_id_as_bytes=tm_data[0:4])
            instance._action_id = struct.unpack("!I", tm_data[4:8])[0]
            instance._custom_data = tm_data[8:]
        else:
            instance.set_packet_info("Unknown functional commanding reply")

    @classmethod
    def __empty(cls) -> Service8FsfwTm:
        return cls(
            subservice=0,
            object_id=bytearray(4),
            action_id=0,
            custom_data=bytearray(),
        )

    @classmethod
    def unpack(cls, raw_telemetry: bytes, time_reader: Optional[CcsdsTimeProvider]):
        service_8_tm = cls.__empty()
        service_8_tm.pus_tm = PusTelemetry.unpack(
            data=raw_telemetry, time_reader=time_reader
        )
        service_8_tm.__init_without_base(instance=service_8_tm)
        return service_8_tm

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(self._object_id.as_hex_string)
        content_list.append(self._action_id)

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Object ID")
        header_list.append("Action ID")

    @property
    def source_object_id_as_bytes(self) -> bytes:
        return bytes(self._object_id.as_bytes)

    @property
    def source_object_id(self) -> UnsignedByteField:
        return self._object_id

    @property
    def action_id(self) -> int:
        return self._action_id

    @property
    def custom_data(self) -> bytearray:
        return self._custom_data



import os.path
from pathlib import Path
import tempfile

from pyfakefs.fake_filesystem_unittest import TestCase
from tmtccmd.cfdp.filestore import HostFilestore, FilestoreResult


class TestCfdpHostFilestore(TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.temp_dir = tempfile.gettempdir()
        self.test_file_name_0 = Path(f"{self.temp_dir}/cfdp_unittest0.txt")
        self.test_file_name_1 = Path(f"{self.temp_dir}/cfdp_unittest1.txt")
        self.test_dir_name_0 = Path(f"{self.temp_dir}/cfdp_test_folder0")
        self.test_dir_name_1 = Path(f"{self.temp_dir}/cfdp_test_folder1")
        self.test_list_dir_name = Path(f"{self.temp_dir}/list-dir-test.txt")
        self.filestore = HostFilestore()

    def test_creation(self):
        res = self.filestore.create_file(self.test_file_name_0)
        self.assertTrue(res == FilestoreResult.CREATE_SUCCESS)
        self.assertTrue(self.test_file_name_0.exists())
        res = self.filestore.create_file(self.test_file_name_0)
        self.assertEqual(res, FilestoreResult.CREATE_NOT_ALLOWED)

        res = self.filestore.delete_file(self.test_file_name_0)
        self.assertEqual(res, FilestoreResult.DELETE_SUCCESS)
        self.assertFalse(os.path.exists(self.test_file_name_0))
        res = self.filestore.delete_file(self.test_file_name_0)
        self.assertTrue(res == FilestoreResult.DELETE_FILE_DOES_NOT_EXIST)

    def test_rename(self):
        self.filestore.create_file(self.test_file_name_0)
        res = self.filestore.rename_file(self.test_file_name_0, self.test_file_name_1)
        self.assertTrue(res == FilestoreResult.RENAME_SUCCESS)
        self.assertTrue(os.path.exists(self.test_file_name_1))
        self.assertFalse(os.path.exists(self.test_file_name_0))
        res = self.filestore.delete_file(self.test_file_name_1)
        self.assertTrue(res == FilestoreResult.DELETE_SUCCESS)

    def test_create_dir(self):
        res = self.filestore.create_directory(self.test_file_name_0)
        self.assertTrue(res == FilestoreResult.CREATE_DIR_SUCCESS)
        self.assertTrue(os.path.isdir(self.test_file_name_0))
        res = self.filestore.create_directory(self.test_file_name_0)
        self.assertTrue(res == FilestoreResult.CREATE_DIR_CAN_NOT_BE_CREATED)

        res = self.filestore.delete_file(self.test_file_name_0)
        self.assertTrue(res == FilestoreResult.DELETE_NOT_ALLOWED)
        res = self.filestore.remove_directory(self.test_file_name_0)
        self.assertTrue(res == FilestoreResult.REMOVE_DIR_SUCCESS)

    def test_read_file(self):
        file_data = "Hello World".encode()
        with open(self.test_file_name_0, "wb") as of:
            of.write(file_data)
        data = self.filestore.read_data(self.test_file_name_0, 0)
        self.assertEqual(data, file_data)

    def test_read_opened_file(self):
        file_data = "Hello World".encode()
        with open(self.test_file_name_0, "wb") as of:
            of.write(file_data)
        with open(self.test_file_name_0, "rb") as rf:
            data = self.filestore.read_from_opened_file(rf, 0, len(file_data))
            self.assertEqual(data, file_data)

    def test_write_file(self):
        file_data = "Hello World".encode()
        self.filestore.create_file(self.test_file_name_0)
        self.filestore.write_data(self.test_file_name_0, file_data, 0)
        with open(self.test_file_name_0, "rb") as rf:
            self.assertEqual(rf.read(), file_data)

    def test_replace_file(self):
        file_data = "Hello World".encode()
        self.filestore.create_file(self.test_file_name_0)
        self.filestore.write_data(self.test_file_name_0, file_data, 0)
        with open(self.test_file_name_1, "w"):
            pass
        self.filestore.replace_file(self.test_file_name_1, self.test_file_name_0)
        self.assertEqual(self.filestore.read_data(self.test_file_name_1, 0), file_data)

    def test_list_dir(self):
        filestore = HostFilestore()
        tempdir = Path(tempfile.gettempdir())
        if os.path.exists(self.test_list_dir_name):
            os.remove(self.test_list_dir_name)
        # Do not delete, user can check file content now
        res = filestore.list_directory(
            dir_name=tempdir, target_file=self.test_list_dir_name
        )
        self.assertTrue(res == FilestoreResult.SUCCESS)

    def tearDown(self):
        pass



from typing import Optional
from abc import abstractmethod, ABC

from tmtccmd.util.obj_id import ObjectIdDictT

from tmtccmd.core import BackendBase
from tmtccmd.util.retval import RetvalDictT

from .com import ComCfgBase, ComInterface
from .tmtc import TmtcDefinitionWrapper
from .defs import default_json_path, CORE_COM_IF_DICT, ComIfDictT


class HookBase(ABC):
    """This hook allows users to adapt the TMTC commander core to the unique mission requirements.
    It is used by implementing all abstract functions and then passing the instance to the
    TMTC commander core.
    """

    def __init__(self, json_cfg_path: Optional[str] = None):
        self.cfg_path = json_cfg_path
        if self.cfg_path is None:
            self.cfg_path = default_json_path()

    @abstractmethod
    def get_object_ids(self) -> ObjectIdDictT:
        from tmtccmd.config.objects import get_core_object_ids

        """The user can specify an object ID dictionary here mapping object ID bytearrays to a
        list. This list could contain containing the string representation or additional
        information about that object ID.
        """
        return get_core_object_ids()

    @abstractmethod
    def assign_communication_interface(self, com_if_key: str) -> Optional[ComInterface]:
        """Assign the communication interface used by the TMTC commander to send and receive
        TMTC with.

        :param com_if_key:      String key of the communication interface to be created.
        """
        from tmtccmd.config.com import create_com_interface_default

        cfg_base = ComCfgBase(com_if_key=com_if_key, json_cfg_path=self.cfg_path)
        return create_com_interface_default(cfg_base)

    def get_com_if_dict(self) -> ComIfDictT:
        return CORE_COM_IF_DICT

    @abstractmethod
    def get_tmtc_definitions(self) -> TmtcDefinitionWrapper:
        """This is a dicitonary mapping services represented by strings to an operation code
        dictionary.

        :return:
        """
        from tmtccmd.config.globals import get_default_tmtc_defs

        return get_default_tmtc_defs()

    def perform_mode_operation(self, tmtc_backend: BackendBase, mode: int):
        """Perform custom mode operations.

        :param tmtc_backend:
        :param mode:
        :return:
        """
        print("No custom mode operation implemented")

    def get_retval_dict(self) -> RetvalDictT:
        from tmtccmd import get_console_logger

        logger = get_console_logger()
        logger.info("No return value dictionary specified")
        return dict()



"""Configuration helpers and definitions.

Submodules:

* :py:mod:`tmtccmd.config.hook` - Base hook class which should be implemented by user and is used
  by the framework to retrieve certain configuration from the user.
* :py:mod:`tmtccmd.config.args` - Various helper methods and classes to create the argument parsers
  and arguments converts to create the data structures expected by this library from passed CLI
  arguments.
"""
from pathlib import Path
from typing import Optional

from spacepackets.util import ByteFieldEmpty
from tmtccmd.core import TmMode, TcMode

from .args import (
    SetupParams,
    create_default_args_parser,
    add_default_tmtccmd_args,
    parse_default_tmtccmd_input_arguments,
    DefaultProcedureParams,
    PreArgsParsingWrapper,
    ProcedureParamsWrapper,
    CfdpParams,
)
from .defs import (
    CoreModeList,
    CoreModeConverter,
    CoreComInterfaces,
    CORE_COM_IF_DICT,
    default_json_path,
    CoreServiceList,
    ComIfDictT,
)
from .prompt import prompt_op_code, prompt_service
from .tmtc import TmtcDefinitionWrapper, OpCodeEntry, OpCodeOptionBase
from .hook import HookBase
from tmtccmd.tc.procedure import (
    DefaultProcedureInfo,
    CfdpProcedureInfo,
    TcProcedureType,
    ProcedureWrapper,
)
from tmtccmd.cfdp.request import PutRequestCfg, PutRequest
from tmtccmd.core.base import ModeWrapper


def backend_mode_conversion(mode: CoreModeList, mode_wrapper: ModeWrapper):
    if mode == CoreModeConverter.get_str(CoreModeList.LISTENER_MODE):
        mode_wrapper.tm_mode = TmMode.LISTENER
        mode_wrapper.tc_mode = TcMode.IDLE
    elif mode == CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE):
        mode_wrapper.tm_mode = TmMode.LISTENER
        mode_wrapper.tc_mode = TcMode.ONE_QUEUE
    elif mode == CoreModeConverter.get_str(CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE):
        mode_wrapper.tc_mode = TcMode.MULTI_QUEUE
        mode_wrapper.tm_mode = TmMode.LISTENER


def get_global_hook_obj() -> Optional[HookBase]:
    """This function can be used to get the handle to the global hook object.
    :return:
    """
    from tmtccmd import get_console_logger

    logger = get_console_logger()
    try:
        from tmtccmd.core.globals_manager import get_global
        from tmtccmd.config.definitions import CoreGlobalIds

        from typing import cast

        hook_obj_raw = get_global(CoreGlobalIds.TMTC_HOOK)
        if hook_obj_raw is None:
            logger.error("Hook object is invalid!")
            return None
        return cast(HookBase, hook_obj_raw)
    except ImportError:
        logger.exception("Issues importing modules to get global hook handle!")
        return None
    except AttributeError:
        logger.exception("Attribute error when trying to get global hook handle!")
        return None


class SetupWrapper:
    """This class encapsulates various important setup parameters required by tmtccmd components"""

    def __init__(
        self,
        hook_obj: HookBase,
        setup_params: SetupParams,
        proc_param_wrapper: ProcedureParamsWrapper,
        json_cfg_path: Optional[str] = None,
    ):
        """
        :param hook_obj: User hook object. Needs to be implemented by the user
        :param setup_params: Optional helper wrapper which contains CLI arguments.
        """
        self.hook_obj = hook_obj
        self.json_cfg_path = json_cfg_path
        self._params = setup_params
        self.proc_param_wrapper = proc_param_wrapper
        self.json_cfg_path = json_cfg_path
        if json_cfg_path is None:
            self.json_cfg_path = default_json_path()

    @property
    def params(self):
        return self._params


def tmtc_params_to_procedure(params: DefaultProcedureParams) -> DefaultProcedureInfo:
    return DefaultProcedureInfo(service=params.service, op_code=params.op_code)


def cfdp_put_req_params_to_procedure(params: CfdpParams) -> CfdpProcedureInfo:
    proc_info = CfdpProcedureInfo()
    put_req_cfg = PutRequestCfg(
        destination_id=ByteFieldEmpty(),
        source_file=Path(params.source),
        dest_file=params.target,
        closure_requested=params.closure_requested,
        trans_mode=params.transmission_mode,
    )
    proc_info.request_wrapper.base = PutRequest(put_req_cfg)
    return proc_info


def params_to_procedure_conversion(
    param_wrapper: ProcedureParamsWrapper,
) -> ProcedureWrapper:
    proc_wrapper = ProcedureWrapper(None)
    if param_wrapper.ptype == TcProcedureType.DEFAULT:
        proc_wrapper.base = tmtc_params_to_procedure(param_wrapper.def_params())
    elif param_wrapper.ptype == TcProcedureType.CFDP:
        proc_wrapper.base = cfdp_put_req_params_to_procedure(
            param_wrapper.cfdp_params()
        )
    return proc_wrapper



"""`sat-rs <https://egit.irs.uni-stuttgart.de/rust/sat-rs.git>`_ specific PUS event support."""
from .s5_satrs_event_defs import *  # noqa re-export
from tmtccmd.tc.pus_5_event import *  # noqa re-export



from pathlib import Path
from typing import Optional, List, cast

from spacepackets.cfdp import (
    SegmentationControl,
    TransmissionMode,
    FaultHandlerOverrideTlv,
    FlowLabelTlv,
    MessageToUserTlv,
    FileStoreRequestTlv,
)
from spacepackets.util import UnsignedByteField
from tmtccmd.cfdp.defs import CfdpRequestType
import dataclasses


class CfdpRequestBase:
    def __init__(self, req_type: CfdpRequestType):
        self.req_type = req_type


@dataclasses.dataclass
class PutRequestCfg:
    destination_id: UnsignedByteField
    # All the following fields are optional because a put request can also be a metadata-only
    # request
    source_file: Optional[Path]
    dest_file: Optional[str]
    trans_mode: Optional[TransmissionMode]
    closure_requested: Optional[bool]
    seg_ctrl: Optional[
        SegmentationControl
    ] = SegmentationControl.NO_RECORD_BOUNDARIES_PRESERVATION
    fault_handler_overrides: Optional[FaultHandlerOverrideTlv] = None
    flow_label_tlv: Optional[FlowLabelTlv] = None
    msgs_to_user: Optional[List[MessageToUserTlv]] = None
    fs_requests: Optional[List[FileStoreRequestTlv]] = None


class PutRequest(CfdpRequestBase):
    def __init__(self, cfg: PutRequestCfg):
        super().__init__(CfdpRequestType.PUT)
        self.cfg = cfg

    def __repr__(self):
        return f"{self.__class__.__name__}(cfg={self.cfg})"

    @property
    def metadata_only(self):
        if self.cfg.source_file is None and self.cfg.dest_file is None:
            return True
        return False

    def __str__(self):
        src_file_str = "Unknown source file"
        dest_file_str = "Unknown destination file"
        if not self.metadata_only:
            src_file_str = f"Source File: {self.cfg.source_file}"
            dest_file_str = f"Destination File: {self.cfg.dest_file}"
        if self.cfg.trans_mode:
            if self.cfg.trans_mode == TransmissionMode.ACKNOWLEDGED:
                trans_mode_str = "Transmission Mode: Class 2 Acknowledged"
            else:
                trans_mode_str = "Transmission Mode: Class 1 Unacknowledged"
        else:
            trans_mode_str = "Transmission Mode from MIB"
        if self.cfg.closure_requested is not None:
            if self.cfg.closure_requested:
                closure_str = "Closure requested"
            else:
                closure_str = "No closure requested"
        else:
            closure_str = "Closure information from MIB"
        if not self.metadata_only:
            print_str = (
                f"Destination ID: {self.cfg.destination_id}\n"
                f"{src_file_str}\n{dest_file_str}\n{trans_mode_str}\n{closure_str}"
            )
        else:
            # TODO: Print out other parameters
            print_str = f"Destination ID: {self.cfg.destination_id}"
        return print_str


class CfdpRequestWrapper:
    def __init__(self, base: Optional[CfdpRequestBase]):
        self.base = base

    @property
    def request_type(self) -> CfdpRequestType:
        return self.base.req_type

    @property
    def request(self) -> CfdpRequestType:
        return self.base.req_type

    def to_put_request(self) -> PutRequest:
        if self.base.req_type != CfdpRequestType.PUT:
            raise TypeError(f"Request is not a {PutRequest.__name__}: {self.base!r}")
        return cast(PutRequest, self.base)



from __future__ import annotations

import enum

from spacepackets.util import UnsignedByteField


class CfdpRequestType(enum.Enum):
    PUT = 0
    REPORT = 1
    CANCEL = 2
    SUSPEND = 3
    RESUME = 4


class CfdpIndication(enum.Enum):
    TRANSACTION = 0
    EOF = 1
    FINISHED = 2
    METADATA = 3
    FILE_SEGMENT_RECV = 4
    REPORT = 5
    SUSPENDED = 6
    RESUMED = 7
    FAULT = 8
    ABANDONED = 9
    EOF_RECV = 10


class CfdpStates(enum.Enum):
    IDLE = 0
    BUSY_CLASS_1_NACKED = 1
    BUSY_CLASS_2_ACKED = 2
    SUSPENDED = 3


class TransactionId:
    def __init__(
        self,
        source_entity_id: UnsignedByteField,
        transaction_seq_num: UnsignedByteField,
    ):
        self.source_id = source_entity_id
        self.seq_num = transaction_seq_num

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(source_entity_id={self.source_id!r}, "
            f"transaction_seq_num={self.seq_num!r})"
        )

    def __str__(self):
        return (
            f"Transaction {{ Source ID {{{self.source_id}}}, Sequence "
            f"number {self.seq_num.value} }}"
        )

    def __eq__(self, other: TransactionId):
        return self.source_id == other.source_id and self.seq_num == other.seq_num

    def __hash__(self):
        return hash((self.source_id, self.seq_num))


class SequenceNumberOverflow(Exception):
    pass



import os
import tempfile
from pathlib import Path
from typing import Optional
from unittest import TestCase
from unittest.mock import MagicMock

from crcmod.predefined import PredefinedCrc

from spacepackets.cfdp import (
    TransmissionMode,
    NULL_CHECKSUM_U32,
    ConditionCode,
    ChecksumType,
)
from spacepackets.cfdp.pdu import DirectiveType, FileDataPdu
from spacepackets.util import ByteFieldU16, UnsignedByteField, ByteFieldU32
from tmtccmd.cfdp import IndicationCfg, LocalEntityCfg, RemoteEntityCfg
from tmtccmd.cfdp.defs import CfdpStates
from tmtccmd.cfdp.handler import SourceHandler, FsmResult
from tmtccmd.cfdp.handler.defs import PacketSendNotConfirmed
from tmtccmd.cfdp.handler.source import TransactionStep
from tmtccmd.cfdp.request import PutRequestCfg, PutRequest, CfdpRequestWrapper
from tmtccmd.util import SeqCountProvider
from .cfdp_fault_handler_mock import FaultHandler
from .cfdp_user_mock import CfdpUser


class TestCfdpSourceHandler(TestCase):
    """It should be noted that this only verifies the correct generation of PDUs. There is
    no reception handler in play here which would be responsible for generating the files
    from these PDUs
    """

    def common_setup(self, closure_requested: bool):
        self.indication_cfg = IndicationCfg(True, True, True, True, True, True)
        self.fault_handler = FaultHandler()
        self.local_cfg = LocalEntityCfg(
            ByteFieldU16(1), self.indication_cfg, self.fault_handler
        )
        self.cfdp_user = CfdpUser()
        self.cfdp_user.eof_sent_indication = MagicMock()
        self.cfdp_user.transaction_indication = MagicMock()
        self.cfdp_user.transaction_finished_indication = MagicMock()
        self.seq_num_provider = SeqCountProvider(bit_width=8)
        self.source_id = ByteFieldU16(1)
        self.dest_id = ByteFieldU16(2)
        self.file_path = Path(f"{tempfile.gettempdir()}/hello.txt")
        if self.file_path.exists():
            os.remove(self.file_path)
        with open(self.file_path, "w"):
            pass
        self.file_segment_len = 256
        self.remote_cfg = RemoteEntityCfg(
            entity_id=self.dest_id,
            max_file_segment_len=self.file_segment_len,
            closure_requested=closure_requested,
            crc_on_transmission=False,
            default_transmission_mode=TransmissionMode.UNACKNOWLEDGED,
            crc_type=ChecksumType.CRC_32,
            check_limit=None,
        )
        # Create an empty file and send it via CFDP
        self.source_handler = SourceHandler(
            self.local_cfg, self.seq_num_provider, self.cfdp_user
        )

    def _common_empty_file_test(self):
        dest_path = "/tmp/hello_copy.txt"
        dest_id = ByteFieldU16(2)
        self.seq_num_provider.get_and_increment = MagicMock(return_value=3)
        put_req_cfg = PutRequestCfg(
            destination_id=dest_id,
            source_file=self.file_path,
            dest_file=dest_path,
            # Let the transmission mode be auto-determined by the remote MIB
            trans_mode=None,
            closure_requested=None,
        )
        self._start_source_transaction(dest_id, PutRequest(put_req_cfg))
        fsm_res = self.source_handler.state_machine()
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.SENDING_EOF
        )
        self.assertEqual(self.source_handler.transaction_seq_num.value, 3)
        self.assertTrue(fsm_res.pdu_holder.is_file_directive)
        self.assertEqual(fsm_res.pdu_holder.pdu_directive_type, DirectiveType.EOF_PDU)
        eof_pdu = fsm_res.pdu_holder.to_eof_pdu()
        self.assertEqual(eof_pdu.transaction_seq_num.value, 3)
        self.assertEqual(eof_pdu.file_checksum, NULL_CHECKSUM_U32)
        self.assertEqual(eof_pdu.file_size, 0)
        self.assertEqual(eof_pdu.condition_code, ConditionCode.NO_ERROR)
        self.assertEqual(eof_pdu.fault_location, None)
        self.source_handler.confirm_packet_sent_advance_fsm()
        # This indication will be called if the EOF send was confirmed
        self.cfdp_user.eof_sent_indication.assert_called_once()
        self.assertEqual(self.cfdp_user.eof_sent_indication.call_count, 1)

    def _common_small_file_test(self, closure_requested: bool, file_content: str):
        dest_path = "/tmp/hello_copy.txt"
        self.source_id = ByteFieldU32(1)
        self.dest_id = ByteFieldU32(2)
        self.seq_num_provider.get_and_increment = MagicMock(return_value=2)
        self.source_handler.source_id = self.source_id
        put_req_cfg = PutRequestCfg(
            destination_id=self.dest_id,
            source_file=self.file_path,
            dest_file=dest_path,
            # Let the transmission mode be auto-determined by the remote MIB
            trans_mode=None,
            closure_requested=closure_requested,
        )
        with open(self.file_path, "wb") as of:
            crc32 = PredefinedCrc("crc32")
            data = file_content.encode()
            crc32.update(data)
            crc32 = crc32.digest()
            of.write(data)
        file_size = self.file_path.stat().st_size
        self._start_source_transaction(self.dest_id, PutRequest(put_req_cfg))
        self.assertEqual(self.source_handler.transaction_seq_num.value, 2)
        fsm_res = self.source_handler.state_machine()
        file_data_pdu = self._check_fsm_and_contained_file_data(fsm_res)
        self.assertFalse(file_data_pdu.has_segment_metadata)
        self.assertEqual(file_data_pdu.transaction_seq_num.value, 2)
        self.assertEqual(file_data_pdu.file_data, "Hello World\n".encode())
        self.assertEqual(file_data_pdu.offset, 0)
        self.source_handler.confirm_packet_sent_advance_fsm()
        fsm_res = self.source_handler.state_machine()
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.SENDING_EOF
        )
        self.assertEqual(fsm_res.pdu_holder.pdu_directive_type, DirectiveType.EOF_PDU)
        eof_pdu = fsm_res.pdu_holder.to_eof_pdu()
        self.assertEqual(crc32, eof_pdu.file_checksum)
        self.assertEqual(eof_pdu.file_size, file_size)
        self.assertEqual(eof_pdu.condition_code, ConditionCode.NO_ERROR)
        with self.assertRaises(PacketSendNotConfirmed):
            self.source_handler.state_machine()

    def _check_fsm_and_contained_file_data(self, fsm_res: FsmResult) -> FileDataPdu:
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.SENDING_FILE_DATA
        )
        self.assertFalse(fsm_res.pdu_holder.is_file_directive)
        return fsm_res.pdu_holder.to_file_data_pdu()

    def _start_source_transaction(
        self, dest_id: UnsignedByteField, put_request: PutRequest
    ):
        wrapper = CfdpRequestWrapper(put_request)
        self.remote_cfg.entity_id = dest_id
        self.source_handler.start_cfdp_transaction(wrapper, self.remote_cfg)
        fsm_res = self.source_handler.state_machine()
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.SENDING_METADATA
        )
        self.cfdp_user.transaction_indication.assert_called_once()
        self.assertEqual(self.cfdp_user.transaction_indication.call_count, 1)
        self.assertTrue(fsm_res.pdu_holder.is_file_directive)
        self.assertEqual(
            fsm_res.pdu_holder.pdu_directive_type, DirectiveType.METADATA_PDU
        )
        metadata_pdu = fsm_res.pdu_holder.to_metadata_pdu()
        if put_request.cfg.closure_requested is not None:
            self.assertEqual(
                metadata_pdu.params.closure_requested, put_request.cfg.closure_requested
            )
        self.assertEqual(metadata_pdu.checksum_type, ChecksumType.CRC_32)
        self.assertEqual(metadata_pdu.source_file_name, self.file_path.as_posix())
        self.assertEqual(metadata_pdu.dest_file_name, put_request.cfg.dest_file)
        self.assertEqual(metadata_pdu.dest_entity_id, dest_id)
        self.source_handler.confirm_packet_sent_advance_fsm()

    def _verify_eof_indication(self):
        self.source_handler.confirm_packet_sent_advance_fsm()
        self.cfdp_user.eof_sent_indication.assert_called_once()
        self.assertEqual(self.cfdp_user.eof_sent_indication.call_count, 1)

    def _state_checker(
        self,
        fsm_res: Optional[FsmResult],
        expected_state: CfdpStates,
        expected_step: TransactionStep,
    ):
        if fsm_res is not None:
            self.assertEqual(fsm_res.states.state, expected_state)
            self.assertEqual(fsm_res.states.step, expected_step)
        self.assertEqual(self.source_handler.states.state, expected_state)
        self.assertEqual(self.source_handler.states.step, expected_step)

    def tearDown(self) -> None:
        if self.file_path.exists():
            os.remove(self.file_path)



from .defs import CfdpIndication, TransactionId
from .request import CfdpRequestWrapper
from .user import CfdpUserBase
from .filestore import HostFilestore
from .mib import (
    LocalEntityCfg,
    RemoteEntityCfgTable,
    RemoteEntityCfg,
    IndicationCfg,
)



from abc import abstractmethod, ABC
from pathlib import Path


class ProvidesSeqCount(ABC):
    @property
    @abstractmethod
    def max_bit_width(self) -> int:
        pass

    @max_bit_width.setter
    @abstractmethod
    def max_bit_width(self, width: int):
        pass

    @abstractmethod
    def get_and_increment(self) -> int:
        """Contract: Retrieve the current sequence count and then increment it. The first call
        should yield 0"""
        raise NotImplementedError(
            "Please use a concrete class implementing this method"
        )

    def __next__(self):
        return self.get_and_increment()


class FileSeqCountProvider(ProvidesSeqCount):
    """Sequence count provider which uses a disk file to store the current sequence count
    in a non-volatile way. The first call with the next built-in or using the base
    class :py:meth:`get` call will yield a 0
    """

    def __init__(self, max_bit_width: int, file_name: Path = Path("seqcnt.txt")):
        self.file_name = file_name
        self._max_bit_width = max_bit_width
        if not self.file_name.exists():
            self.create_new()

    @property
    def max_bit_width(self) -> int:
        return self._max_bit_width

    @max_bit_width.setter
    def max_bit_width(self, width: int):
        self._max_bit_width = width

    def create_new(self):
        with open(self.file_name, "w") as file:
            file.write("0\n")

    def current(self) -> int:
        if not self.file_name.exists():
            raise FileNotFoundError(f"{self.file_name} file does not exist")
        with open(self.file_name) as file:
            return self.check_count(file.readline())

    def get_and_increment(self) -> int:
        if not self.file_name.exists():
            raise FileNotFoundError(f"{self.file_name} file does not exist")
        with open(self.file_name, "r+") as file:
            curr_seq_cnt = self.check_count(file.readline())
            file.seek(0)
            file.write(f"{self.increment_with_rollover(curr_seq_cnt)}\n")
            return curr_seq_cnt

    def check_count(self, line: str) -> int:
        line = line.rstrip()
        if not line.isdigit():
            raise ValueError("Sequence count file content is invalid")
        curr_seq_cnt = int(line)
        if curr_seq_cnt < 0 or curr_seq_cnt > pow(2, self.max_bit_width) - 1:
            raise ValueError("Sequence count in file has invalid value")
        return curr_seq_cnt

    def increment_with_rollover(self, seq_cnt: int) -> int:
        """CCSDS Sequence count has maximum size of 14 bit. Rollover after that size by default"""
        if seq_cnt >= pow(2, self.max_bit_width) - 1:
            return 0
        else:
            return seq_cnt + 1


class PusFileSeqCountProvider(FileSeqCountProvider):
    def __init__(self, file_name: Path = Path("seqcnt.txt")):
        super().__init__(max_bit_width=14, file_name=file_name)


class SeqCountProvider(ProvidesSeqCount):
    def __init__(self, bit_width: int):
        self.count = 0
        self._max_bit_width = bit_width

    @property
    def max_bit_width(self) -> int:
        return self._max_bit_width

    @max_bit_width.setter
    def max_bit_width(self, width: int):
        self._max_bit_width = width

    def get_and_increment(self) -> int:
        curr_count = self.count
        self.count += 1
        return curr_count



"""`FSFW <https://egit.irs.uni-stuttgart.de/fsfw/fsfw>`_ specific PUS event support"""
from .s5_fsfw_event_defs import *  # noqa re-export
from tmtccmd.tm.pus_5_fsfw_event import *  # noqa re-export
from tmtccmd.tc.pus_5_event import *  # noqa re-export



import logging

from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.shortcuts import CompleteStyle
import prompt_toolkit
from tmtccmd.config.tmtc import OpCodeEntry, TmtcDefinitionWrapper

_LOGGER = logging.getLogger(__name__)


def prompt_service(
    tmtc_defs: TmtcDefinitionWrapper,
    compl_style: CompleteStyle = CompleteStyle.READLINE_LIKE,
) -> str:
    service_adjustment = 20
    info_adjustment = 30
    horiz_line_num = service_adjustment + info_adjustment + 3
    horiz_line = horiz_line_num * "-"
    service_ladjusted = "Service".ljust(service_adjustment)
    info_string = "Information".ljust(info_adjustment)
    tmtc_defs.sort()
    while True:
        print(f" {horiz_line}")
        print(f"|{service_ladjusted} | {info_string}|")
        print(f" {horiz_line}")
        srv_completer = build_service_word_completer(tmtc_defs)
        for service_entry in tmtc_defs.defs.items():
            try:
                adjusted_service_entry = service_entry[0].ljust(service_adjustment)
                adjusted_service_info = service_entry[1][0].ljust(info_adjustment)
                print(f"|{adjusted_service_entry} | {adjusted_service_info}|")
            except AttributeError:
                _LOGGER.warning(
                    f"Error handling service entry {service_entry[0]}. Skipping.."
                )
        print(f" {horiz_line}")
        service_string = prompt_toolkit.prompt(
            "Please select a service by specifying the key: ",
            completer=srv_completer,
            complete_style=compl_style,
        )
        if service_string in tmtc_defs.defs:
            print(f"Selected service: {service_string}")
            return service_string
        else:
            _LOGGER.warning("Invalid key, try again")


def build_service_word_completer(
    tmtc_defs: TmtcDefinitionWrapper,
) -> WordCompleter:
    srv_list = []
    for service_entry in tmtc_defs.defs.items():
        srv_list.append(service_entry[0])
    srv_completer = WordCompleter(words=srv_list, ignore_case=True)
    return srv_completer


def prompt_op_code(
    tmtc_defs: TmtcDefinitionWrapper,
    service: str,
    compl_style: CompleteStyle = CompleteStyle.READLINE_LIKE,
) -> str:
    op_code_adjustment = 24
    info_adjustment = 56
    horz_line_num = op_code_adjustment + info_adjustment + 3
    horiz_line = horz_line_num * "-"
    op_code_info_number_str = "Operation Code (Number)".ljust(op_code_adjustment)
    op_code_info_str_str = "Operation Code (Text)".ljust(op_code_adjustment)
    info_string = "Information".ljust(info_adjustment)
    if service in tmtc_defs.defs:
        op_code_entry = tmtc_defs.op_code_entry(service)
        completer = build_op_code_word_completer(op_code_entry=op_code_entry)
    else:
        _LOGGER.warning("Service not in dictionary. Setting default operation code 0")
        return "0"
    while True:

        def print_table(otype: str, dictionary: dict):
            print(f" {horiz_line}")
            print(f"|{otype} | {info_string}|")
            print(f" {horiz_line}")
            for op_code in dictionary.items():
                adjusted_op_code_entry = op_code[0].ljust(op_code_adjustment)
                adjusted_op_code_info = op_code[1][0].ljust(info_adjustment)
                print(f"|{adjusted_op_code_entry} | {adjusted_op_code_info}|")
            print(f" {horiz_line}")

        if len(op_code_entry.op_code_dict_str_keys) > 0:
            print_table(op_code_info_str_str, op_code_entry.op_code_dict_str_keys)
        if len(op_code_entry.op_code_dict_num_keys) > 0:
            print_table(op_code_info_number_str, op_code_entry.op_code_dict_num_keys)
        op_code_string = prompt_toolkit.prompt(
            "Please select an operation code by specifying the key: ",
            completer=completer,
            complete_style=compl_style,
        )
        if (
            op_code_string in op_code_entry.op_code_dict_str_keys.keys()
            or op_code_string in op_code_entry.op_code_dict_num_keys.keys()
        ):
            print(f"Selected op code: {op_code_string}")
            return op_code_string
        else:
            _LOGGER.warning("Invalid key, try again")


def build_op_code_word_completer(op_code_entry: OpCodeEntry) -> WordCompleter:
    op_code_list = []
    for op_code_str in op_code_entry.op_code_dict_num_keys.keys():
        op_code_list.append(op_code_str)
    for op_code_str in op_code_entry.op_code_dict_str_keys.keys():
        op_code_list.append(op_code_str)
    op_code_completer = WordCompleter(words=op_code_list, ignore_case=True)
    return op_code_completer



from pathlib import Path
from typing import Optional

from crcmod.predefined import PredefinedCrc

from spacepackets.cfdp import ChecksumType, NULL_CHECKSUM_U32
from tmtccmd.cfdp.filestore import VirtualFilestore
from tmtccmd.cfdp.handler.defs import ChecksumNotImplemented, SourceFileDoesNotExist


class Crc32Helper:
    def __init__(self, init_type: ChecksumType, vfs: VirtualFilestore):
        self.checksum_type = init_type
        self.vfs = vfs

    def _verify_checksum(self):
        if self.checksum_type not in [
            ChecksumType.NULL_CHECKSUM,
            ChecksumType.CRC_32,
            ChecksumType.CRC_32C,
        ]:
            raise ChecksumNotImplemented(self.checksum_type)

    def checksum_type_to_crcmod_str(self) -> Optional[str]:
        if self.checksum_type == ChecksumType.NULL_CHECKSUM:
            return None
        if self.checksum_type == ChecksumType.CRC_32:
            return "crc32"
        elif self.checksum_type == ChecksumType.CRC_32C:
            return "crc32c"

    def generate_crc_calculator(self) -> PredefinedCrc:
        self._verify_checksum()
        return PredefinedCrc(self.checksum_type_to_crcmod_str())

    def calc_for_file(self, file: Path, file_sz: int, segment_len: int) -> bytes:
        if self.checksum_type == ChecksumType.NULL_CHECKSUM:
            return NULL_CHECKSUM_U32
        crc_obj = self.generate_crc_calculator()
        if segment_len == 0:
            raise ValueError("Segment length can not be 0")
        if not file.exists():
            raise SourceFileDoesNotExist(file)
        current_offset = 0
        # Calculate the file CRC
        with open(file, "rb") as of:
            while current_offset < file_sz:
                if current_offset + segment_len > file_sz:
                    read_len = file_sz - current_offset
                else:
                    read_len = segment_len
                if read_len > 0:
                    crc_obj.update(
                        self.vfs.read_from_opened_file(of, current_offset, read_len)
                    )
                current_offset += read_len
            return crc_obj.digest()



import time
from typing import Optional

from PyQt5.QtCore import QRunnable, pyqtSlot, QObject, pyqtSignal

from tmtccmd import get_console_logger
from tmtccmd.core import TmMode, TcMode, BackendRequest
from tmtccmd.gui.defs import LocalArgs, SharedArgs, WorkerOperationsCodes


LOGGER = get_console_logger()


class WorkerSignalWrapper(QObject):
    finished = pyqtSignal(object)
    failure = pyqtSignal(object)
    stop = pyqtSignal(object)
    abort = pyqtSignal(object)


class FrontendWorker(QRunnable):
    """Runnable thread which can be used with QThreadPool. Not used for now, might be needed
    in the future.
    """

    def __init__(self, local_args: LocalArgs, shared_args: SharedArgs):
        super(QRunnable, self).__init__()
        self._locals = local_args
        self._shared = shared_args
        self.signals = WorkerSignalWrapper()
        self._stop_signal = False
        self._abort_signal = False
        self.signals.stop.connect(self._stop_com_if)
        self.signals.abort.connect(self._abort)

    def __sanitize_locals(self):
        if self._locals.op_code == WorkerOperationsCodes.LISTEN_FOR_TM:
            if self._locals.op_args is None or not isinstance(
                float, self._locals.op_args
            ):
                self._locals.op_args = 0.2

    def __setup(self, op_code: WorkerOperationsCodes) -> bool:
        if op_code == WorkerOperationsCodes.OPEN_COM_IF:
            LOGGER.info("Switching COM Interface")
            # TODO: We should really pass a proper object here instead of using magic tuples..
            new_com_if = self._locals.op_args[2].assign_communication_interface(
                com_if_key=self._locals.op_args[1]
            )
            # self._args.state.last_com_if = self._args.state.current_com_if
            set_success = self._shared.backend.try_set_com_if(new_com_if)
            if not set_success:
                LOGGER.warning(
                    f"Could not set new communication interface {new_com_if}"
                )
            if self._shared.backend.com_if_active():
                self._finish_with_info("COM Interface is already active")
            else:
                self._shared.backend.open_com_if()
                self._finish_success()
            return False
        if op_code == WorkerOperationsCodes.CLOSE_COM_IF:
            if not self._shared.backend.com_if_active():
                self._finish_with_info("COM Interface is not active")
            elif self._shared.com_if_ref_tracker.is_used():
                self._failure_with_info("Can not close COM interface which is used")
            else:
                self._shared.backend.close_com_if()
                self._finish_success()
            return False
        if op_code == WorkerOperationsCodes.ONE_QUEUE_MODE:
            self._shared.com_if_ref_tracker.add_user()
            with self._shared.tc_lock:
                self._shared.backend.tc_mode = TcMode.ONE_QUEUE
        elif op_code == WorkerOperationsCodes.LISTEN_FOR_TM:
            self._shared.com_if_ref_tracker.add_user()
            self._shared.backend.tm_mode = TmMode.LISTENER
        return True

    def __one_queue_mode_cycle(self) -> Optional[bool]:
        self._shared.tc_lock.acquire()
        self._shared.backend.tc_operation()
        self._update_backend_mode()
        state = self._shared.backend.state
        if state.request == BackendRequest.TERMINATION_NO_ERROR:
            self._shared.tc_lock.release()
            self._shared.com_if_ref_tracker.remove_user()
            with self._shared.state_lock:
                if (
                    not self._shared.com_if_ref_tracker.is_used()
                    and self._locals.op_args is not None
                ):
                    self._locals.op_args()
            self._finish_success()
            return False
        elif state.request == BackendRequest.DELAY_IDLE:
            time.sleep(1.0)
        elif state.request == BackendRequest.DELAY_CUSTOM:
            self._shared.tc_lock.release()
            if state.next_delay.total_seconds() < 0.5:
                time.sleep(state.next_delay.total_seconds())
            else:
                time.sleep(0.5)
        elif state.request == BackendRequest.CALL_NEXT:
            self._shared.tc_lock.release()

    def __listener_cycle(self) -> Optional[bool]:
        if self._stop_signal or self._abort_signal:
            self._shared.com_if_ref_tracker.remove_user()
            if not self._abort_signal:
                self._finish_success()
            return False
        else:
            # We only should run the TM operation here
            self._shared.backend.tm_operation()
            # Poll TM every 400 ms for now
            time.sleep(self._locals.op_args)

    def __loop(self, op_code: WorkerOperationsCodes) -> bool:
        if op_code == WorkerOperationsCodes.ONE_QUEUE_MODE:
            should_return = self.__one_queue_mode_cycle()
            if should_return is not None:
                return should_return
        elif op_code == WorkerOperationsCodes.LISTEN_FOR_TM:
            should_return = self.__listener_cycle()
            if should_return is not None:
                return should_return
        elif op_code == WorkerOperationsCodes.IDLE:
            return False
        else:
            # This must be a programming error
            get_console_logger().error(
                f"Unknown worker operation code {self._locals.op_code}"
            )
        return True

    @pyqtSlot()
    def run(self):
        op_code = self._locals.op_code
        loop_required = self.__setup(op_code)
        if loop_required:
            while True:
                if not self.__loop(op_code):
                    break

    def _finish_success(self):
        self.signals.finished.emit(None)

    def _finish_with_info(self, info: str):
        self.signals.finished.emit(info)

    def _failure_with_info(self, info: str):
        self.signals.failure.emit(info)

    def _update_backend_mode(self):
        with self._shared.state_lock:
            self._shared.backend.mode_to_req()

    def _stop_com_if(self, _args: any):
        self._stop_signal = True

    def _abort(self, _args: any):
        self._abort_signal = True



from __future__ import annotations

import logging
import struct
from typing import Optional

from spacepackets.ccsds.time import CcsdsTimeProvider
from spacepackets.ecss.defs import PusService
from spacepackets.ecss.tm import CdsShortTimestamp, PusTelemetry

from tmtccmd.tm.base import PusTmInfoBase, PusTmBase


class FileInfo:
    def __init__(
        self,
        file_name: str,
        repo_path: str,
        file_size: int = 0,
        lock_status: bool = False,
    ):
        self.file_name = file_name
        self.repo_path = repo_path
        self.file_size = file_size
        self.lock_status = lock_status


class Service23Tm(PusTmInfoBase, PusTmBase):
    MAX_REPOSITORY_LENGTH = 64
    MAX_FILENAME_LENGTH = 12

    def __init__(
        self,
        subservice_id: int,
        object_id: bytearray,
        repo_path: str,
        file_name: str,
        time: Optional[CdsShortTimestamp],
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        self.object_id = object_id
        self.data_start_idx = 0
        self.file_info = FileInfo(file_name=file_name, repo_path=repo_path)
        pus_tm = PusTelemetry(
            service=PusService.S23_FILE_MGMT,
            subservice=subservice_id,
            time_provider=time,
            seq_count=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)

        self.set_packet_info("File Service Reply")

    @staticmethod
    def __init_without_base(instance: Service23Tm):
        tm_data = instance.tm_data
        if len(tm_data) < 4:
            logging.getLogger(__name__).error("Service23TM: Invalid packet format!")
            return
        instance.object_id = struct.unpack("!I", tm_data[0:4])[0]
        instance.file_info.file_size = 0
        instance.file_info.lock_status = False
        instance.data_start_idx = 0
        if instance.subservice == 4:
            instance.unpack_repo_and_filename()
            instance.unpack_file_attributes()
        elif instance.subservice == 132:
            instance.unpack_repo_and_filename()

    @classmethod
    def __empty(cls) -> Service23Tm:
        return cls(
            subservice_id=0,
            object_id=bytearray(4),
            repo_path="",
            file_name="",
            time=CdsShortTimestamp.empty(),
        )

    @classmethod
    def unpack(
        cls, raw_telemetry: bytes, time_reader: Optional[CcsdsTimeProvider]
    ) -> Service23Tm:
        service_23_tm = cls.__empty()
        service_23_tm.pus_tm = PusTelemetry.unpack(
            data=raw_telemetry, time_reader=time_reader
        )
        service_23_tm.__init_without_base(instance=service_23_tm)
        return service_23_tm

    def unpack_repo_and_filename(self):
        tm_data = self.tm_data
        repo_path_found = False
        path_idx_start = 0
        max_len_to_scan = len(self.tm_data) - 4
        for idx in range(4, max_len_to_scan):
            if not repo_path_found and tm_data[idx] == 0:
                repo_bytes = tm_data[4:idx]
                self.file_info.repo_path = repo_bytes.decode("utf-8")
                path_idx_start = idx + 1
                idx += 1
                repo_path_found = True
            if repo_path_found:
                if tm_data[idx] == 0:
                    filename_bytes = tm_data[path_idx_start:idx]
                    self.file_info.file_name = filename_bytes.decode("utf-8")
                    self.data_start_idx = idx + 1
                    break

    def unpack_file_attributes(self):
        # Size of file length (4) + lock status (1), adapt if more field are added!
        print(len(self.tm_data) - self.data_start_idx)
        if len(self.tm_data) - self.data_start_idx != 5:
            logging.getLogger(__name__).error(
                "Service23TM: Invalid lenght of file attributes data"
            )
            return
        self.file_info.file_size = struct.unpack(
            "!I", self.tm_data[self.data_start_idx : self.data_start_idx + 4]
        )[0]
        self.file_info.lock_status = self.tm_data[self.data_start_idx + 4]

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list)
        content_list.append(self.object_id)
        content_list.append(self.file_info.repo_path)
        content_list.append(self.file_info.file_name)
        if self.subservice == 4:
            content_list.append(self.file_info.file_size)
            if self.file_info.lock_status == 0:
                content_list.append("No")
            else:
                content_list.append("Yes")

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list)
        header_list.append("Object ID")
        header_list.append("Repo Path")
        header_list.append("File Name")
        if self.subservice == 4:
            header_list.append("File Size")
            header_list.append("Locked")
